# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
QEMU VM implementation.
"""

from gns3.vm import VM
from gns3.node import Node
from gns3.ports.port import Port
from gns3.ports.ethernet_port import EthernetPort
from .settings import QEMU_VM_SETTINGS

import logging
log = logging.getLogger(__name__)


class QemuVM(VM):

    """
    QEMU VM.

    :param module: parent module for this node
    :param server: GNS3 server instance
    """
    URL_PREFIX = "qemu"

    def __init__(self, module, server, project):
        VM.__init__(self, module, server, project)

        log.info("QEMU VM instance is being created")
        self._defaults = {}
        self._inital_settings = None
        self._export_directory = None
        self._loading = False
        self._module = module
        self._ports = []

        self._settings = {"name": "",
                          "qemu_path": "",
                          "hda_disk_image": "",
                          "hdb_disk_image": "",
                          "hdc_disk_image": "",
                          "hdd_disk_image": "",
                          "options": "",
                          "ram": QEMU_VM_SETTINGS["ram"],
                          "console": None,
                          "adapters": QEMU_VM_SETTINGS["adapters"],
                          "adapter_type": QEMU_VM_SETTINGS["adapter_type"],
                          "legacy_networking": QEMU_VM_SETTINGS["legacy_networking"],
                          "cpu_throttling": QEMU_VM_SETTINGS["cpu_throttling"],
                          "process_priority": QEMU_VM_SETTINGS["process_priority"],
                          "initrd": "",
                          "kernel_image": "",
                          "kernel_command_line": ""}

        self._addAdapters(1)

        # save the default settings
        self._defaults = self._settings.copy()

    def _addAdapters(self, adapters):
        """
        Adds adapters.

        :param adapters: number of adapters
        """

        for adapter_number in range(0, adapters):
            adapter_name = EthernetPort.longNameType() + str(adapter_number)
            short_name = EthernetPort.shortNameType() + str(adapter_number)
            new_port = EthernetPort(adapter_name)
            new_port.setShortName(short_name)
            new_port.setAdapterNumber(adapter_number)
            new_port.setPortNumber(0)
            self._ports.append(new_port)
            log.debug("Adapter {} has been added".format(adapter_name))

    def setup(self, qemu_path, name=None, console=None, vm_id=None, initial_settings={}, base_name=None):
        """
        Setups this QEMU VM.

        :param name: optional name
        :param vm_id: VM identifier
        """

        # let's create a unique name if none has been chosen
        if not name:
            name = self.allocateName(base_name + "-")

        if not name:
            self.error_signal.emit(self.id(), "could not allocate a name for this QEMU VM")
            return

        self._settings["name"] = name
        params = {"name": name,
                  "qemu_path": qemu_path}

        if console:
            params["console"] = self._settings["console"] = console

        if vm_id:
            params["vm_id"] = vm_id

        # other initial settings will be applied when the router has been created
        if initial_settings:
            self._inital_settings = initial_settings

        self.httpPost("/qemu/vms", self._setupCallback, body=params)

    def _setupCallback(self, result, error=False, **kwargs):
        """
        Callback for setup.

        :param result: server response
        :param error: indicates an error (boolean)
        """

        if error:
            log.error("error while setting up {}: {}".format(self.name(), result["message"]))
            self.server_error_signal.emit(self.id(), result["message"])
            return

        self._vm_id = result["vm_id"]
        if not self._vm_id:
            self.error_signal.emit(self.id(), "returned ID from server is null")
            return

        # update the settings using the defaults sent by the server
        for name, value in result.items():
            if name in self._settings and self._settings[name] != value:
                log.info("QEMU VM instance {} setting up and updating {} from '{}' to '{}'".format(self.name(),
                                                                                                   name,
                                                                                                   self._settings[name],
                                                                                                   value))
                self._settings[name] = value

        # update the node with setup initial settings if any
        if self._inital_settings:
            self.update(self._inital_settings)
        elif self._loading:
            self.updated_signal.emit()
        else:
            self.setInitialized(True)
            log.info("QEMU VM instance {} has been created".format(self.name()))
            self.created_signal.emit(self.id())
            self._module.addNode(self)

    def delete(self):
        """
        Deletes this QEMU VM instance.
        """

        log.debug("QEMU VM instance {} is being deleted".format(self.name()))
        # first delete all the links attached to this node
        self.delete_links_signal.emit()
        if self._vm_id:
            self.httpDelete("/qemu/vms/{vm_id}".format(vm_id=self._vm_id), self._deleteCallback)
        else:
            self.deleted_signal.emit()
            self._module.removeNode(self)

    def _deleteCallback(self, result, error=False, **kwargs):
        """
        Callback for delete.

        :param result: server response
        :param error: indicates an error (boolean)
        """

        if error:
            log.error("error while deleting {}: {}".format(self.name(), result["message"]))
            self.server_error_signal.emit(self.id(), result["message"])
        log.info("{} has been deleted".format(self.name()))
        self.deleted_signal.emit()
        self._module.removeNode(self)

    def update(self, new_settings):
        """
        Updates the settings for this QEMU VM.

        :param new_settings: settings dictionary
        """

        if "name" in new_settings and new_settings["name"] != self.name() and self.hasAllocatedName(new_settings["name"]):
            self.error_signal.emit(self.id(), 'Name "{}" is already used by another node'.format(new_settings["name"]))
            return

        params = {}
        for name, value in new_settings.items():
            if name in self._settings and self._settings[name] != value:
                params[name] = value

        if "cloud_path" in new_settings:
            params["cloud_path"] = self._settings["cloud_path"] = new_settings.pop("cloud_path")

        log.debug("{} is updating settings: {}".format(self.name(), params))
        self.httpPut("/qemu/vms/{vm_id}".format(vm_id=self._vm_id), self._updateCallback, body=params)

    def _updateCallback(self, result, error=False, **kwargs):
        """
        Callback for update.

        :param result: server response
        :param error: indicates an error (boolean)
        """

        if error:
            log.error("error while deleting {}: {}".format(self.name(), result["message"]))
            self.server_error_signal.emit(self.id(), result["message"])
            return

        updated = False
        nb_adapters_changed = False
        for name, value in result.items():
            if name in self._settings and self._settings[name] != value:
                log.info("{}: updating {} from '{}' to '{}'".format(self.name(), name, self._settings[name], value))
                updated = True
                if name == "name":
                    # update the node name
                    self.updateAllocatedName(value)
                if name == "adapters":
                    nb_adapters_changed = True
                self._settings[name] = value

        if nb_adapters_changed:
            log.debug("number of adapters has changed to {}".format(self._settings["adapters"]))
            # TODO: dynamically add/remove adapters
            self._ports.clear()
            self._addAdapters(self._settings["adapters"])

        if self._inital_settings and not self._loading:
            self.setInitialized(True)
            log.info("QEMU VM {} has been created".format(self.name()))
            self.created_signal.emit(self.id())
            self._module.addNode(self)
            self._inital_settings = None
        elif updated or self._loading:
            log.info("QEMU VM {} has been updated".format(self.name()))
            self.updated_signal.emit()

    def start(self):
        """
        Starts this QEMU VM instance.
        """

        if self.status() == Node.started:
            log.debug("{} is already running".format(self.name()))
            return

        log.debug("{} is starting".format(self.name()))
        self.httpPost("/qemu/vms/{vm_id}/start".format(vm_id=self._vm_id), self._startCallback)

    def _startCallback(self, result, error=False, **kwargs):
        """
        Callback for start.

        :param result: server response
        :param error: indicates an error (boolean)
        """

        if error:
            log.error("error while starting {}: {}".format(self.name(), result["message"]))
            self.server_error_signal.emit(self.id(), result["message"])
        else:
            log.info("{} has started".format(self.name()))
            self.setStatus(Node.started)
            for port in self._ports:
                # set ports as started
                port.setStatus(Port.started)
            self.started_signal.emit()

    def stop(self):
        """
        Stops this QEMU VM instance.
        """

        if self.status() == Node.stopped:
            log.debug("{} is already stopped".format(self.name()))
            return

        log.debug("{} is stopping".format(self.name()))
        self.httpPost("/qemu/vms/{vm_id}/stop".format(vm_id=self._vm_id), self._stopCallback)

    def _stopCallback(self, result, error=False, **kwargs):
        """
        Callback for stop.

        :param result: server response
        :param error: indicates an error (boolean)
        """

        if error:
            log.error("error while stopping {}: {}".format(self.name(), result["message"]))
            self.server_error_signal.emit(self.id(), result["message"])
        else:
            log.info("{} has stopped".format(self.name()))
            self.setStatus(Node.stopped)
            for port in self._ports:
                # set ports as stopped
                port.setStatus(Port.stopped)
            self.stopped_signal.emit()

    def suspend(self):
        """
        Suspends this QEMU VM instance.
        """

        if self.status() == Node.suspended:
            log.debug("{} is already suspended".format(self.name()))
            return

        log.debug("{} is being suspended".format(self.name()))
        self.httpPost("/qemu/vms/{vm_id}/suspend".format(vm_id=self._vm_id), self._suspendCallback)

    def _suspendCallback(self, result, error=False, **kwargs):
        """
        Callback for suspend.

        :param result: server response
        :param error: indicates an error (boolean)
        """

        if error:
            log.error("error while suspending {}: {}".format(self.name(), result["message"]))
            self.server_error_signal.emit(self.id(), result["message"])
        else:
            log.info("{} has suspended".format(self.name()))
            self.setStatus(Node.suspended)
            for port in self._ports:
                # set ports as suspended
                port.setStatus(Port.suspended)
            self.suspended_signal.emit()

    def reload(self):
        """
        Reloads this QEMU VM instance.
        """

        log.debug("{} is being reloaded".format(self.name()))
        self.httpPost("/qemu/vms/{vm_id}/reload".format(vm_id=self._vm_id), self._reloadCallback)

    def _reloadCallback(self, result, error=False, **kwargs):
        """
        Callback for reload.

        :param result: server response
        :param error: indicates an error (boolean)
        """

        if error:
            log.error("error while reloading {}: {}".format(self.name(), result["message"]))
            self.server_error_signal.emit(self.id(), result["message"])
        else:
            log.info("{} has reloaded".format(self.name()))

    def dump(self):
        """
        Returns a representation of this QEMU VM instance.
        (to be saved in a topology file).

        :returns: representation of the node (dictionary)
        """

        qemu_vm = {"id": self.id(),
                   "vm_id": self._vm_id,
                   "type": self.__class__.__name__,
                   "description": str(self),
                   "properties": {},
                   "server_id": self._server.id()}

        # add the properties
        for name, value in self._settings.items():
            if name in self._defaults and self._defaults[name] != value:
                qemu_vm["properties"][name] = value

        # add the ports
        if self._ports:
            ports = qemu_vm["ports"] = []
            for port in self._ports:
                ports.append(port.dump())

        return qemu_vm

    def info(self):
        """
        Returns information about this QEMU VM instance.

        :returns: formated string
        """

        if self.status() == Node.started:
            state = "started"
        else:
            state = "stopped"

        info = """QEMU VM {name} is {state}
  Node ID is {id}, server's QEMU VM ID is {vm_id}
  QEMU VM's server runs on {host}:{port}, console is on port {console}
""".format(name=self.name(),
           id=self.id(),
           vm_id=self._vm_id,
           state=state,
           host=self._server.host,
           port=self._server.port,
           console=self._settings["console"])

        port_info = ""
        for port in self._ports:
            if port.isFree():
                port_info += "     {port_name} is empty\n".format(port_name=port.name())
            else:
                port_info += "     {port_name} {port_description}\n".format(port_name=port.name(),
                                                                            port_description=port.description())

        return info + port_info

    def load(self, node_info):
        """
        Loads a QEMU VM representation
        (from a topology file).

        :param node_info: representation of the node (dictionary)
        """

        self.node_info = node_info
        # for backward compatibility
        vm_id = node_info.get("qemu_id")
        if not vm_id:
            vm_id = node_info["vm_id"]
        settings = node_info["properties"]
        name = settings.pop("name")
        qemu_path = settings.pop("qemu_path")
        console = settings.pop("console", self._defaults["console"])
        self.updated_signal.connect(self._updatePortSettings)
        # block the created signal, it will be triggered when loading is completely done
        self._loading = True
        log.info("QEMU VM {} is loading".format(name))
        self.setName(name)
        self.setup(qemu_path, name, console, vm_id, settings)

    def _updatePortSettings(self):
        """
        Updates port settings when loading a topology.
        """

        self.updated_signal.disconnect(self._updatePortSettings)
        # update the port with the correct names and IDs
        if "ports" in self.node_info:
            ports = self.node_info["ports"]
            for topology_port in ports:
                for port in self._ports:
                    adapter_number = topology_port.get("adapter_number", topology_port["port_number"])
                    if adapter_number == port.adapterNumber():
                        port.setName(topology_port["name"])
                        port.setId(topology_port["id"])

        # now we can set the node has initialized and trigger the signal
        self.setInitialized(True)
        log.info("QEMU VM {} has been loaded".format(self.name()))
        self.created_signal.emit(self.id())
        self._module.addNode(self)
        self._inital_settings = None
        self._loading = False

    def name(self):
        """
        Returns the name of this QEMU VM instance.

        :returns: name (string)
        """

        return self._settings["name"]

    def settings(self):
        """
        Returns all this QEMU VM instance settings.

        :returns: settings dictionary
        """

        return self._settings

    def ports(self):
        """
        Returns all the ports for this QEMU VM instance.

        :returns: list of Port instances
        """

        return self._ports

    def console(self):
        """
        Returns the console port for this QEMU VM instance.

        :returns: port (integer)
        """

        return self._settings["console"]

    def configPage(self):
        """
        Returns the configuration page widget to be used by the node configurator.

        :returns: QWidget object
        """

        from .pages.qemu_vm_configuration_page import QemuVMConfigurationPage
        return QemuVMConfigurationPage

    @staticmethod
    def defaultSymbol():
        """
        Returns the default symbol path for this node.

        :returns: symbol path (or resource).
        """

        return ":/symbols/qemu_guest.normal.svg"

    @staticmethod
    def hoverSymbol():
        """
        Returns the symbol to use when this node is hovered.

        :returns: symbol path (or resource).
        """

        return ":/symbols/qemu_guest.selected.svg"

    @staticmethod
    def symbolName():

        return "QEMU VM"

    @staticmethod
    def categories():
        """
        Returns the node categories the node is part of (used by the device panel).

        :returns: list of node category (integer)
        """

        return [Node.end_devices]

    def __str__(self):

        return "QEMU VM"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Base cloud controller class.

Base class for interacting with Cloud APIs to create and manage cloud
instances.

"""
from collections import namedtuple
import hashlib
import os
import logging
from io import StringIO, BytesIO

from libcloud.compute.base import NodeAuthSSHKey
from libcloud.storage.types import ContainerAlreadyExistsError, ContainerDoesNotExistError, ObjectDoesNotExistError

from .exceptions import ItemNotFound, KeyPairExists, MethodNotAllowed
from .exceptions import OverLimit, BadRequest, ServiceUnavailable
from .exceptions import Unauthorized, ApiError


KeyPair = namedtuple("KeyPair", ['name'], verbose=False)
log = logging.getLogger(__name__)


def parse_exception(exception):
    """
    Parse the exception to separate the HTTP status code from the text.

    Libcloud raises many exceptions of the form:
        Exception("<http status code> <http error> <reponse body>")

    in lieu of raising specific incident-based exceptions.

    """

    e_str = str(exception)

    try:
        status = int(e_str[0:3])
        error_text = e_str[3:]

    except ValueError:
        status = None
        error_text = e_str

    return status, error_text


class BaseCloudCtrl(object):

    """ Base class for interacting with a cloud provider API. """

    http_status_to_exception = {
        400: BadRequest,
        401: Unauthorized,
        404: ItemNotFound,
        405: MethodNotAllowed,
        413: OverLimit,
        500: ApiError,
        503: ServiceUnavailable
    }

    GNS3_CONTAINER_NAME = 'GNS3'

    def __init__(self, username, api_key):
        self.username = username
        self.api_key = api_key

    def _handle_exception(self, status, error_text, response_overrides=None):
        """ Raise an exception based on the HTTP status. """

        if response_overrides:
            if status in response_overrides:
                raise response_overrides[status](error_text)

        raise self.http_status_to_exception[status](error_text)

    def authenticate(self):
        """ Validate cloud account credentials.  Return boolean. """
        raise NotImplementedError

    def list_sizes(self):
        """ Return a list of NodeSize objects. """

        return self.driver.list_sizes()

    def list_flavors(self):
        """ Return an iterable of flavors """

        raise NotImplementedError

    def create_instance(self, name, size_id, image_id, keypair):
        """
        Create a new instance with the supplied attributes.

        Return a Node object.

        """
        try:
            image = self.get_image(image_id)
            if image is None:
                raise ItemNotFound("Image not found")

            size = self.driver.ex_get_size(size_id)

            args = {
                "name": name,
                "size": size,
                "image": image,
            }

            if keypair is not None:
                auth_key = NodeAuthSSHKey(keypair.public_key)
                args["auth"] = auth_key
                args["ex_keyname"] = name

            return self.driver.create_node(**args)

        except Exception as e:
            status, error_text = parse_exception(e)

            if status:
                self._handle_exception(status, error_text)
            else:
                log.error("create_instance method raised an exception: {}".format(e))
                log.error('image id {}'.format(image_id))

    def delete_instance(self, instance):
        """ Delete the specified instance.  Returns True or False. """

        try:
            return self.driver.destroy_node(instance)

        except Exception as e:

            status, error_text = parse_exception(e)

            if status:
                self._handle_exception(status, error_text)
            else:
                raise e

    def get_instance(self, instance):
        """ Return a Node object representing the requested instance. """

        for i in self.driver.list_nodes():
            if i.id == instance.id:
                return i

        raise ItemNotFound("Instance not found")

    def list_instances(self):
        """ Return a list of instances in the current region. """

        return self.driver.list_nodes()

    def create_key_pair(self, name):
        """ Create and return a new Key Pair. """

        response_overrides = {
            409: KeyPairExists
        }
        try:
            return self.driver.create_key_pair(name)

        except Exception as e:
            status, error_text = parse_exception(e)
            if status:
                self._handle_exception(status, error_text, response_overrides)
            else:
                raise e

    def delete_key_pair(self, keypair):
        """ Delete the keypair. Returns True or False. """

        try:
            return self.driver.delete_key_pair(keypair)

        except Exception as e:
            status, error_text = parse_exception(e)
            if status:
                self._handle_exception(status, error_text)
            else:
                raise e

    def delete_key_pair_by_name(self, keypair_name):
        """ Utility method to incapsulate boilerplate code """

        kp = KeyPair(name=keypair_name)
        return self.delete_key_pair(kp)

    def list_key_pairs(self):
        """ Return a list of Key Pairs. """

        return self.driver.list_key_pairs()

    def upload_file(self, file_path, cloud_object_name):
        """
        Uploads file to cloud storage (if it is not identical to a file already in cloud storage).
        :param file_path: path to file to upload
        :param cloud_object_name: name of file saved in cloud storage
        :return: True if file was uploaded, False if it was skipped because it already existed and was identical
        """
        try:
            gns3_container = self.storage_driver.create_container(self.GNS3_CONTAINER_NAME)
        except ContainerAlreadyExistsError:
            gns3_container = self.storage_driver.get_container(self.GNS3_CONTAINER_NAME)

        with open(file_path, 'rb') as file:
            local_file_hash = hashlib.md5(file.read()).hexdigest()

            cloud_hash_name = cloud_object_name + '.md5'
            cloud_objects = [obj.name for obj in gns3_container.list_objects()]

            # if the file and its hash are in object storage, and the local and storage file hashes match
            # do not upload the file, otherwise upload it
            if cloud_object_name in cloud_objects and cloud_hash_name in cloud_objects:
                hash_object = gns3_container.get_object(cloud_hash_name)
                cloud_object_hash = ''
                for chunk in hash_object.as_stream():
                    cloud_object_hash += chunk.decode('utf8')

                if cloud_object_hash == local_file_hash:
                    return False

            file.seek(0)
            self.storage_driver.upload_object_via_stream(file, gns3_container, cloud_object_name)
            self.storage_driver.upload_object_via_stream(StringIO(local_file_hash), gns3_container, cloud_hash_name)
            return True

    def list_projects(self):
        """
        Lists projects in cloud storage
        :return: Dictionary where project names are keys and values are names of objects in storage
        """

        try:
            gns3_container = self.storage_driver.get_container(self.GNS3_CONTAINER_NAME)
            projects = {
                obj.name.replace('projects/', '').replace('.zip', ''): obj.name
                for obj in gns3_container.list_objects()
                if obj.name.startswith('projects/') and obj.name[-4:] == '.zip'
            }
            return projects
        except ContainerDoesNotExistError:
            return []

    def download_file(self, file_name, destination=None):
        """
        Downloads file from cloud storage. If a file exists at destination, and it is identical to the file in cloud
        storage, it is not downloaded.
        :param file_name: name of file in cloud storage to download
        :param destination: local path to save file to (if None, returns file contents as a file-like object)
        :return: A file-like object if file contents are returned, or None if file is saved to filesystem
        """

        gns3_container = self.storage_driver.get_container(self.GNS3_CONTAINER_NAME)
        storage_object = gns3_container.get_object(file_name)

        if destination is not None:
            if os.path.isfile(destination):
                # if a file exists at destination and its hash matches that of the
                # file in cloud storage, don't download it
                with open(destination, 'rb') as f:
                    local_file_hash = hashlib.md5(f.read()).hexdigest()

                    hash_object = gns3_container.get_object(file_name + '.md5')
                    cloud_object_hash = ''
                    for chunk in hash_object.as_stream():
                        cloud_object_hash += chunk.decode('utf8')

                    if local_file_hash == cloud_object_hash:
                        return

            storage_object.download(destination)
        else:
            contents = b''

            for chunk in storage_object.as_stream():
                contents += chunk

            return BytesIO(contents)

    def find_storage_image_names(self, images_to_find):
        """
        Maps names of image files to their full name in cloud storage
        :param images_to_find: list of image names to find
        :return: A dictionary where keys are image names, and values are the corresponding names of
        the files in cloud storage
        """
        gns3_container = self.storage_driver.get_container(self.GNS3_CONTAINER_NAME)
        images_in_storage = [obj.name for obj in gns3_container.list_objects() if obj.name.startswith('images/')]

        images = {}
        for image_name in images_to_find:
            images_with_same_name =\
                list(filter(lambda storage_image_name: storage_image_name.endswith(image_name), images_in_storage))

            if len(images_with_same_name) == 1:
                images[image_name] = images_with_same_name[0]
            else:
                raise Exception('Image does not exist in cloud storage or is duplicated')

        return images

    def delete_file(self, file_name):
        gns3_container = self.storage_driver.get_container(self.GNS3_CONTAINER_NAME)

        try:
            object_to_delete = gns3_container.get_object(file_name)
            object_to_delete.delete()
        except ObjectDoesNotExistError:
            pass

        try:
            hash_object = gns3_container.get_object(file_name + '.md5')
            hash_object.delete()
        except ObjectDoesNotExistError:
            pass



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from ..qt import QtGui
from ..topology import Topology
from ..ui.idlepc_dialog_ui import Ui_IdlePCDialog


class IdlePCDialog(QtGui.QDialog, Ui_IdlePCDialog):

    """
    Idle-PC dialog.
    """

    def __init__(self, router, idlepcs, parent):

        QtGui.QDialog.__init__(self, parent)
        self.setupUi(self)
        self.uiButtonBox.button(QtGui.QDialogButtonBox.Apply).clicked.connect(self._applySlot)
        self.uiButtonBox.button(QtGui.QDialogButtonBox.Help).clicked.connect(self._helpSlot)

        self._router = router
        self._idlepcs = idlepcs

        for value in self._idlepcs:
            match = re.search(r"^(0x[0-9a-f]+)\s+\[(\d+)\]$", value)
            if match:
                idlepc = match.group(1)
                count = int(match.group(2))
                if 50 <= count <= 60:
                    value += "*"
                self.uiComboBox.addItem(value, idlepc)

    def _helpSlot(self):
        """
        Shows the help for Idle-PC.
        """

        help_text = "Finding the right idlepc value is a trial and error process, consisting of applying " \
                    "different Idle-PC values and monitoring the CPU usage.\n\nBest Idle-PC values are usually " \
                    "obtained when IOS is in idle state, the following message being displayed " \
                    "on the console: {} con0 is now available ... Press RETURN to get started.".format(self._router.name())
        QtGui.QMessageBox.information(self, "Hints for Idle-PC", help_text)

    def _applySlot(self):
        """
        Applies an Idle-PC value.
        """

        if not self.uiComboBox.count():
            QtGui.QMessageBox.critical(self, "Idle-PC", "Sorry could not find a valid Idle-PC value, please check again with Cisco IOS in a different state")
            return

        idlepc = self.uiComboBox.itemData(self.uiComboBox.currentIndex())
        # apply Idle-PC to all routers with the same IOS image
        ios_image = self._router.settings()["image"]
        for node in Topology.instance().nodes():
            if hasattr(node, "idlepc") and node.settings()["image"] == ios_image:
                node.setIdlepc(idlepc)

    def done(self, result):
        """
        Called when the dialog is closed.

        :param result: boolean (accepted or rejected)
        """

        if result:
            self._applySlot()
        QtGui.QDialog.done(self, result)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

from ..qt import QtCore, QtGui, QtWebKit
from ..ui.getting_started_dialog_ui import Ui_GettingStartedDialog
from ..utils.get_resource import get_resource
from ..local_config import LocalConfig


class GettingStartedDialog(QtGui.QDialog, Ui_GettingStartedDialog):

    """
    GettingStarted dialog.
    """

    def __init__(self, parent):

        QtGui.QDialog.__init__(self, parent)
        self.setupUi(self)

        self.uiWebView.page().mainFrame().setScrollBarPolicy(QtCore.Qt.Horizontal, QtCore.Qt.ScrollBarAlwaysOff)
        self.uiWebView.page().mainFrame().setScrollBarPolicy(QtCore.Qt.Vertical, QtCore.Qt.ScrollBarAlwaysOff)
        self.adjustSize()
        self.uiWebView.page().setLinkDelegationPolicy(QtWebKit.QWebPage.DelegateAllLinks)
        self.uiWebView.linkClicked.connect(self._urlClickedSlot)
        self._local_config = LocalConfig.instance()
        gui_settings = self._local_config.loadSectionSettings("GUI", {"hide_getting_started_dialog": False})
        self.uiCheckBox.setChecked(gui_settings["hide_getting_started_dialog"])
        getting_started = get_resource(os.path.join("static", "getting_started.html"))
        if getting_started and not (sys.platform.startswith("win") and not sys.maxsize > 2 ** 32):
            # do not show the page on Windows 32-bit (crash when no Internet connection)
            self.uiWebView.load(QtCore.QUrl.fromLocalFile(getting_started))
        else:
            self.uiCheckBox.setChecked(True)
            self.accept()

    def showit(self):
        """
        Either this dialog should be automatically showed at startup.

        :returns: boolean
        """

        return not self.uiCheckBox.isChecked()

    def done(self, result):
        """
        This dialog is closed.

        :param result: ignored
        """

        self._local_config.saveSectionSettings("GUI", {"hide_getting_started_dialog": self.uiCheckBox.isChecked()})
        QtGui.QDialog.done(self, result)

    def _urlClickedSlot(self, url):
        """
        Opens a clicked URL using user's default browser.

        :param url: URL to open
        """

        if QtGui.QDesktopServices.openUrl(url) is False:
            QtGui.QMessageBox.critical(self, "Getting started", "Failed to open the URL: {}".format(url))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gns3.node import Node
from .cloud import Cloud

import logging
log = logging.getLogger(__name__)


class Host(Cloud):

    """
    Pseudo host based on a Dynamips Cloud.

    :param module: parent module for this node
    :param server: GNS3 server instance
    :param project: Project instance
    """

    _name_instance_count = 1

    def __init__(self, module, server, project):
        Cloud.__init__(self, module, server, project)

        log.info("host is being created")
        # create an unique id and name
        self._name_id = Host._name_instance_count
        Host._name_instance_count += 1

        name = "Host{}".format(self._name_id)
        self._settings["name"] = name

    def setup(self, name=None, initial_settings={}):
        """
        Setups this host.

        :param name: optional name for this host
        """

        if name:
            self._settings["name"] = name

        if initial_settings:
            self._initial_settings = initial_settings
        else:
            self.created_signal.connect(self._autoConfigure)

        self._server.get("/interfaces", self._setupCallback)

    def _autoConfigure(self, node_id):
        """
        Auto adds all Ethernet and TAP interfaces.

        :param node_id: ignored
        """

        new_settings = {"nios": []}
        for interface in self._settings["interfaces"]:
            if interface["name"].startswith("tap"):
                new_settings["nios"].append("nio_tap:{}".format(interface["name"]))
            else:
                new_settings["nios"].append("nio_gen_eth:{}".format(interface["name"]))

        self.update(new_settings)

    @staticmethod
    def defaultSymbol():
        """
        Returns the default symbol path for this host.

        :returns: symbol path (or resource).
        """

        return ":/symbols/computer.normal.svg"

    @staticmethod
    def hoverSymbol():
        """
        Returns the symbol to use when the host is hovered.

        :returns: symbol path (or resource).
        """

        return ":/symbols/computer.selected.svg"

    @staticmethod
    def symbolName():

        return "Host"

    @staticmethod
    def categories():
        """
        Returns the node categories the node is part of (used by the device panel).

        :returns: list of node category (integer)
        """

        return [Node.end_devices]

    def __str__(self):

        return "Host"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Configuration page for VPCS devices.
"""

import os

from gns3.qt import QtGui
from gns3.utils.get_resource import get_resource
from ..ui.vpcs_device_configuration_page_ui import Ui_VPCSDeviceConfigPageWidget


class VPCSDeviceConfigurationPage(QtGui.QWidget, Ui_VPCSDeviceConfigPageWidget):

    """
    QWidget configuration page for VPCS devices.
    """

    def __init__(self):

        QtGui.QWidget.__init__(self)
        self.setupUi(self)
        # self.uiScriptFileToolButton.clicked.connect(self._scriptFileBrowserSlot)

    # def _scriptFileBrowserSlot(self):
    #     """
    #     Slot to open a file browser and select a script-file file.
    #     """
    #
    #     config_dir = get_resource("configs")
    #     path = QtGui.QFileDialog.getOpenFileName(self, "Select a startup configuration", config_dir)
    #     if not path:
    #         return
    #
    #     if not os.access(path, os.R_OK):
    #         QtGui.QMessageBox.critical(self, "Startup configuration", "Cannot read {}".format(path))
    #         return
    #
    #     self.uiScriptFileLineEdit.clear()
    #     self.uiScriptFileLineEdit.setText(path)

    def loadSettings(self, settings, node, group=False):
        """
        Loads the VPCS device settings.

        :param settings: the settings (dictionary)
        :param node: Node instance
        :param group: indicates the settings apply to a group of routers
        """

        if not group:
            self.uiNameLineEdit.setText(settings["name"])
            self.uiConsolePortSpinBox.setValue(settings["console"])

            # load the script-file
            # self.uiScriptFileLineEdit.setText(settings["script_file"])

        else:
            self.uiNameLabel.hide()
            self.uiNameLineEdit.hide()
            self.uiConsolePortLabel.hide()
            self.uiConsolePortSpinBox.hide()
            # self.uiScriptFileLabel.hide()
            # self.uiScriptFileLineEdit.hide()
            # self.uiScriptFileToolButton.hide()

    def saveSettings(self, settings, node, group=False):
        """
        Saves the VPCS device settings.

        :param settings: the settings (dictionary)
        :param node: Node instance
        :param group: indicates the settings apply to a group of routers
        """

        # these settings cannot be shared by nodes and updated
        # in the node configurator.
        if not group:

            # set the device name
            name = self.uiNameLineEdit.text()
            if not name:
                QtGui.QMessageBox.critical(self, "Name", "VPCS device name cannot be empty!")
            else:
                settings["name"] = name

            settings["console"] = self.uiConsolePortSpinBox.value()

            # script_file = self.uiScriptFileLineEdit.text()
            # if script_file != settings["script_file"]:
            #    if os.access(script_file, os.R_OK):
            #        settings["script_file"] = script_file
            #    else:
            #        QtGui.QMessageBox.critical(self, "Script-file", "Cannot read the script-file file")
        else:
            del settings["name"]
            del settings["console"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
VPCS module implementation.
"""

import os
import subprocess
import sys
import signal
import socket
import re
import shutil
import pkg_resources

from gns3.servers import Servers
from gns3.local_config import LocalConfig
from gns3.utils.get_resource import get_resource
from gns3.utils.get_default_base_config import get_default_base_config
from gns3.local_server_config import LocalServerConfig
from gns3.qt import QtCore

from ..module import Module
from ..module_error import ModuleError
from .vpcs_device import VPCSDevice
from .settings import VPCS_SETTINGS, VPCS_SETTING_TYPES
from ...settings import ENABLE_CLOUD

import logging
log = logging.getLogger(__name__)


class VPCS(Module):

    """
    VPCS module.
    """

    def __init__(self):
        Module.__init__(self)

        self._settings = {}
        self._nodes = []
        self._working_dir = ""

        self._vpcs_multi_host_process = None
        self._vpcs_multi_host_port = 0

        # load the settings
        self._loadSettings()

    @staticmethod
    def _findVPCS(self):
        """
        Finds the VPCS path.

        :return: path to VPCS
        """

        if sys.platform.startswith("win") and hasattr(sys, "frozen"):
            vpcs_path = os.path.join(os.getcwd(), "vpcs", "vpcs.exe")
        elif sys.platform.startswith("darwin") and hasattr(sys, "frozen"):
            vpcs_path = os.path.join(os.getcwd(), "vpcs")
        else:
            vpcs_path = shutil.which("vpcs")

        if vpcs_path is None:
            return ""
        return vpcs_path

    def _loadSettings(self):
        """
        Loads the settings from the persistent settings file.
        """

        local_config = LocalConfig.instance()

        # restore the VPCS settings from QSettings (for backward compatibility)
        legacy_settings = {}
        settings = QtCore.QSettings()
        settings.beginGroup(self.__class__.__name__)
        for name in VPCS_SETTINGS.keys():
            if settings.contains(name):
                legacy_settings[name] = settings.value(name, type=VPCS_SETTING_TYPES[name])
        settings.remove("")
        settings.endGroup()

        if legacy_settings:
            local_config.saveSectionSettings(self.__class__.__name__, legacy_settings)
        self._settings = local_config.loadSectionSettings(self.__class__.__name__, VPCS_SETTINGS)

        if not self._settings["base_script_file"]:
            self._settings["base_script_file"] = get_default_base_config(get_resource(os.path.join("configs", "vpcs_base_config.txt")))

        if not os.path.exists(self._settings["vpcs_path"]):
            self._settings["vpcs_path"] = self._findVPCS(self)

        # keep the config file sync
        self._saveSettings()

    def _saveSettings(self):
        """
        Saves the settings to the persistent settings file.
        """

        # save the settings
        LocalConfig.instance().saveSectionSettings(self.__class__.__name__, self._settings)

        if self._settings["vpcs_path"]:
            # save some settings to the server config file
            server_settings = {
                "vpcs_path": os.path.normpath(self._settings["vpcs_path"]),
            }
            config = LocalServerConfig.instance()
            config.saveSettings(self.__class__.__name__, server_settings)

    def addNode(self, node):
        """
        Adds a node to this module.

        :param node: Node instance
        """

        self._nodes.append(node)

    def removeNode(self, node):
        """
        Removes a node from this module.

        :param node: Node instance
        """

        if node in self._nodes:
            self._nodes.remove(node)

    def settings(self):
        """
        Returns the module settings

        :returns: module settings (dictionary)
        """

        return self._settings

    def setSettings(self, settings):
        """
        Sets the module settings

        :param settings: module settings (dictionary)
        """

        self._settings.update(settings)
        self._saveSettings()

    def createNode(self, node_class, server, project):
        """
        Creates a new node.

        :param node_class: Node object
        :param server: HTTPClient instance
        :param project: Project instance
        """

        log.info("Creating node {}".format(node_class))

        # create an instance of the node class
        return node_class(self, server, project)

    def setupNode(self, node, node_name):
        """
        Setups a node.

        :param node: Node instance
        :param node_name: Node name
        """

        log.info("configuring node {}".format(node))
        settings = {}

        script_file = self._settings["base_script_file"]
        if script_file:
            settings["script_file"] = script_file

        node.setup(None, additional_settings=settings)

    def reset(self):
        """
        Resets the module.
        """

        log.info("VPCS module reset")
        self._nodes.clear()

    def exportConfigs(self, directory):
        """
        Exports all configs for all nodes to a directory.

        :param directory: destination directory path
        """

        for node in self._nodes:
            if node.initialized():
                node.exportConfigToDirectory(directory)

    def importConfigs(self, directory):
        """
        Imports configs to all nodes from a directory.

        :param directory: source directory path
        """

        for node in self._nodes:
            if node.initialized():
                node.importConfigFromDirectory(directory)

    def _check_vpcs_version(self, working_dir):
        """
        Checks if the VPCS executable version is >= 0.5b1.
        """

        try:
            output = subprocess.check_output([self._settings["vpcs_path"], "-v"], cwd=working_dir)
            match = re.search("Welcome to Virtual PC Simulator, version ([0-9a-z\.]+)", output.decode("utf-8"))
            if match:
                version = match.group(1)
                if pkg_resources.parse_version(version) < pkg_resources.parse_version("0.5b1"):
                    raise ModuleError("VPCS executable version must be >= 0.5b1")
            else:
                raise ModuleError("Could not determine the VPCS version for {}".format(self._settings["vpcs_path"]))
        except (OSError, subprocess.SubprocessError) as e:
            raise ModuleError("Error while looking for the VPCS version: {}".format(e))

    def startMultiHostVPCS(self, working_dir):
        """
        Starts a VPCS process for multi-host support.

        :param working_dir: VPCS multi-host working directory
        :return: VPCS listening port
        """

        if self._vpcs_multi_host_process and self._vpcs_multi_host_process.poll() is None:
            return self._vpcs_multi_host_port

        if not self._settings["vpcs_path"]:
            raise ModuleError("No path to a VPCS executable has been set")

        if not os.path.isfile(self._settings["vpcs_path"]):
            raise ModuleError("VPCS program '{}' is not accessible".format(self._settings["vpcs_path"]))

        if not os.access(self._settings["vpcs_path"], os.X_OK):
            raise ModuleError("VPCS program '{}' is not executable".format(self._settings["vpcs_path"]))

        self._check_vpcs_version(working_dir)

        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.bind(("127.0.0.1", 0))
                self._vpcs_multi_host_port = sock.getsockname()[1]
        except OSError as e:
            raise ModuleError("Cannot get a free port: {}".format(e))

        flags = 0
        if sys.platform.startswith("win32"):
            flags = subprocess.CREATE_NEW_PROCESS_GROUP
        try:
            vpcs_command = [self._settings["vpcs_path"], "-p", str(self._vpcs_multi_host_port), "-F"]
            self._vpcs_multi_host_process = subprocess.Popen(vpcs_command, cwd=working_dir, creationflags=flags)
        except (OSError, subprocess.SubprocessError) as e:
            raise ModuleError("Could not start VPCS {}".format(e))

        return self._vpcs_multi_host_port

    def stopMultiHostVPCS(self):
        """
        Stops the VPCS process for multi-host support.
        """

        if self._vpcs_multi_host_process and self._vpcs_multi_host_process.poll() is None:
            log.info("stopping VPCS multi-host instance PID={}".format(self._vpcs_multi_host_process.pid))
            if sys.platform.startswith("win32"):
                self._vpcs_multi_host_process.send_signal(signal.CTRL_BREAK_EVENT)
            else:
                self._vpcs_multi_host_process.terminate()

            self._vpcs_multi_host_process.wait()

        self._vpcs_multi_host_process = None

    @staticmethod
    def getNodeClass(name):
        """
        Returns the object with the corresponding name.

        :param name: object name
        """

        if name in globals():
            return globals()[name]
        return None

    @staticmethod
    def classes():
        """
        Returns all the node classes supported by this module.

        :returns: list of classes
        """

        return [VPCSDevice]

    def nodes(self):
        """
        Returns all the node data necessary to represent a node
        in the nodes view and create a node on the scene.
        """

        server = "local"
        if not self._settings["use_local_server"]:
            # pick up a remote server (round-robin method)
            remote_server = next(iter(Servers.instance()))
            if remote_server:
                server = "{}:{}".format(remote_server.host, remote_server.port)

        nodes = []
        for node_class in VPCS.classes():
            nodes.append(
                {"class": node_class.__name__,
                 "name": node_class.symbolName(),
                 "server": server,
                 "categories": node_class.categories(),
                 "default_symbol": node_class.defaultSymbol(),
                 "hover_symbol": node_class.hoverSymbol()}
            )
            if ENABLE_CLOUD:
                nodes.append(
                    {"class": node_class.__name__,
                     "name": node_class.symbolName() + " (cloud)",
                     "server": "cloud",
                     "categories": node_class.categories(),
                     "default_symbol": node_class.defaultSymbol(),
                     "hover_symbol": node_class.hoverSymbol()}
                )
        return nodes

    @staticmethod
    def preferencePages():
        """
        :returns: QWidget object list
        """

        from .pages.vpcs_preferences_page import VPCSPreferencesPage
        return [VPCSPreferencesPage]

    @staticmethod
    def instance():
        """
        Singleton to return only on instance of VPCS module.

        :returns: instance of VPCS
        """

        if not hasattr(VPCS, "_instance"):
            VPCS._instance = VPCS()
        return VPCS._instance



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/grossmj/PycharmProjects/gns3-gui/gns3/modules/qemu/ui/qemu_vm_configuration_page.ui'
#
# Created: Mon Mar 23 22:22:18 2015
#      by: PyQt4 UI code generator 4.10.4
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)

class Ui_QemuVMConfigPageWidget(object):
    def setupUi(self, QemuVMConfigPageWidget):
        QemuVMConfigPageWidget.setObjectName(_fromUtf8("QemuVMConfigPageWidget"))
        QemuVMConfigPageWidget.resize(486, 407)
        self.verticalLayout = QtGui.QVBoxLayout(QemuVMConfigPageWidget)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.uiQemutabWidget = QtGui.QTabWidget(QemuVMConfigPageWidget)
        self.uiQemutabWidget.setObjectName(_fromUtf8("uiQemutabWidget"))
        self.tab = QtGui.QWidget()
        self.tab.setObjectName(_fromUtf8("tab"))
        self.gridLayout_4 = QtGui.QGridLayout(self.tab)
        self.gridLayout_4.setObjectName(_fromUtf8("gridLayout_4"))
        self.uiQemuListLabel = QtGui.QLabel(self.tab)
        self.uiQemuListLabel.setObjectName(_fromUtf8("uiQemuListLabel"))
        self.gridLayout_4.addWidget(self.uiQemuListLabel, 2, 0, 1, 2)
        self.uiNameLineEdit = QtGui.QLineEdit(self.tab)
        self.uiNameLineEdit.setObjectName(_fromUtf8("uiNameLineEdit"))
        self.gridLayout_4.addWidget(self.uiNameLineEdit, 0, 2, 1, 1)
        spacerItem = QtGui.QSpacerItem(263, 94, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.gridLayout_4.addItem(spacerItem, 4, 1, 1, 2)
        self.uiQemuListComboBox = QtGui.QComboBox(self.tab)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uiQemuListComboBox.sizePolicy().hasHeightForWidth())
        self.uiQemuListComboBox.setSizePolicy(sizePolicy)
        self.uiQemuListComboBox.setObjectName(_fromUtf8("uiQemuListComboBox"))
        self.gridLayout_4.addWidget(self.uiQemuListComboBox, 2, 2, 1, 1)
        self.uiConsolePortSpinBox = QtGui.QSpinBox(self.tab)
        self.uiConsolePortSpinBox.setMaximum(65535)
        self.uiConsolePortSpinBox.setObjectName(_fromUtf8("uiConsolePortSpinBox"))
        self.gridLayout_4.addWidget(self.uiConsolePortSpinBox, 3, 2, 1, 1)
        self.uiConsolePortLabel = QtGui.QLabel(self.tab)
        self.uiConsolePortLabel.setObjectName(_fromUtf8("uiConsolePortLabel"))
        self.gridLayout_4.addWidget(self.uiConsolePortLabel, 3, 0, 1, 2)
        self.uiNameLabel = QtGui.QLabel(self.tab)
        self.uiNameLabel.setObjectName(_fromUtf8("uiNameLabel"))
        self.gridLayout_4.addWidget(self.uiNameLabel, 0, 0, 1, 1)
        self.uiRamLabel = QtGui.QLabel(self.tab)
        self.uiRamLabel.setObjectName(_fromUtf8("uiRamLabel"))
        self.gridLayout_4.addWidget(self.uiRamLabel, 1, 0, 1, 1)
        self.uiRamSpinBox = QtGui.QSpinBox(self.tab)
        self.uiRamSpinBox.setMinimum(32)
        self.uiRamSpinBox.setMaximum(65535)
        self.uiRamSpinBox.setProperty("value", 256)
        self.uiRamSpinBox.setObjectName(_fromUtf8("uiRamSpinBox"))
        self.gridLayout_4.addWidget(self.uiRamSpinBox, 1, 2, 1, 1)
        self.uiQemutabWidget.addTab(self.tab, _fromUtf8(""))
        self.tab_3 = QtGui.QWidget()
        self.tab_3.setObjectName(_fromUtf8("tab_3"))
        self.gridLayout_6 = QtGui.QGridLayout(self.tab_3)
        self.gridLayout_6.setObjectName(_fromUtf8("gridLayout_6"))
        self.uiHdaDiskImageLabel = QtGui.QLabel(self.tab_3)
        self.uiHdaDiskImageLabel.setObjectName(_fromUtf8("uiHdaDiskImageLabel"))
        self.gridLayout_6.addWidget(self.uiHdaDiskImageLabel, 0, 0, 1, 1)
        self.horizontalLayout_8 = QtGui.QHBoxLayout()
        self.horizontalLayout_8.setObjectName(_fromUtf8("horizontalLayout_8"))
        self.uiHdaDiskImageLineEdit = QtGui.QLineEdit(self.tab_3)
        self.uiHdaDiskImageLineEdit.setObjectName(_fromUtf8("uiHdaDiskImageLineEdit"))
        self.horizontalLayout_8.addWidget(self.uiHdaDiskImageLineEdit)
        self.uiHdaDiskImageToolButton = QtGui.QToolButton(self.tab_3)
        self.uiHdaDiskImageToolButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.uiHdaDiskImageToolButton.setObjectName(_fromUtf8("uiHdaDiskImageToolButton"))
        self.horizontalLayout_8.addWidget(self.uiHdaDiskImageToolButton)
        self.gridLayout_6.addLayout(self.horizontalLayout_8, 0, 1, 1, 1)
        self.uiHdbDiskImageLabel = QtGui.QLabel(self.tab_3)
        self.uiHdbDiskImageLabel.setObjectName(_fromUtf8("uiHdbDiskImageLabel"))
        self.gridLayout_6.addWidget(self.uiHdbDiskImageLabel, 1, 0, 1, 1)
        self.horizontalLayout_4 = QtGui.QHBoxLayout()
        self.horizontalLayout_4.setObjectName(_fromUtf8("horizontalLayout_4"))
        self.uiHdbDiskImageLineEdit = QtGui.QLineEdit(self.tab_3)
        self.uiHdbDiskImageLineEdit.setObjectName(_fromUtf8("uiHdbDiskImageLineEdit"))
        self.horizontalLayout_4.addWidget(self.uiHdbDiskImageLineEdit)
        self.uiHdbDiskImageToolButton = QtGui.QToolButton(self.tab_3)
        self.uiHdbDiskImageToolButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.uiHdbDiskImageToolButton.setObjectName(_fromUtf8("uiHdbDiskImageToolButton"))
        self.horizontalLayout_4.addWidget(self.uiHdbDiskImageToolButton)
        self.gridLayout_6.addLayout(self.horizontalLayout_4, 1, 1, 1, 1)
        self.uiHdcDiskImageLabel = QtGui.QLabel(self.tab_3)
        self.uiHdcDiskImageLabel.setObjectName(_fromUtf8("uiHdcDiskImageLabel"))
        self.gridLayout_6.addWidget(self.uiHdcDiskImageLabel, 2, 0, 1, 1)
        self.horizontalLayout_9 = QtGui.QHBoxLayout()
        self.horizontalLayout_9.setObjectName(_fromUtf8("horizontalLayout_9"))
        self.uiHdcDiskImageLineEdit = QtGui.QLineEdit(self.tab_3)
        self.uiHdcDiskImageLineEdit.setObjectName(_fromUtf8("uiHdcDiskImageLineEdit"))
        self.horizontalLayout_9.addWidget(self.uiHdcDiskImageLineEdit)
        self.uiHdcDiskImageToolButton = QtGui.QToolButton(self.tab_3)
        self.uiHdcDiskImageToolButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.uiHdcDiskImageToolButton.setObjectName(_fromUtf8("uiHdcDiskImageToolButton"))
        self.horizontalLayout_9.addWidget(self.uiHdcDiskImageToolButton)
        self.gridLayout_6.addLayout(self.horizontalLayout_9, 2, 1, 1, 1)
        self.uiHddDiskImageLabel = QtGui.QLabel(self.tab_3)
        self.uiHddDiskImageLabel.setObjectName(_fromUtf8("uiHddDiskImageLabel"))
        self.gridLayout_6.addWidget(self.uiHddDiskImageLabel, 3, 0, 1, 1)
        self.horizontalLayout_10 = QtGui.QHBoxLayout()
        self.horizontalLayout_10.setObjectName(_fromUtf8("horizontalLayout_10"))
        self.uiHddDiskImageLineEdit = QtGui.QLineEdit(self.tab_3)
        self.uiHddDiskImageLineEdit.setObjectName(_fromUtf8("uiHddDiskImageLineEdit"))
        self.horizontalLayout_10.addWidget(self.uiHddDiskImageLineEdit)
        self.uiHddDiskImageToolButton = QtGui.QToolButton(self.tab_3)
        self.uiHddDiskImageToolButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.uiHddDiskImageToolButton.setObjectName(_fromUtf8("uiHddDiskImageToolButton"))
        self.horizontalLayout_10.addWidget(self.uiHddDiskImageToolButton)
        self.gridLayout_6.addLayout(self.horizontalLayout_10, 3, 1, 1, 1)
        spacerItem1 = QtGui.QSpacerItem(438, 257, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.gridLayout_6.addItem(spacerItem1, 4, 0, 1, 2)
        self.uiQemutabWidget.addTab(self.tab_3, _fromUtf8(""))
        self.tab_7 = QtGui.QWidget()
        self.tab_7.setObjectName(_fromUtf8("tab_7"))
        self.gridLayout_5 = QtGui.QGridLayout(self.tab_7)
        self.gridLayout_5.setObjectName(_fromUtf8("gridLayout_5"))
        self.uiAdaptersLabel = QtGui.QLabel(self.tab_7)
        self.uiAdaptersLabel.setObjectName(_fromUtf8("uiAdaptersLabel"))
        self.gridLayout_5.addWidget(self.uiAdaptersLabel, 0, 0, 1, 1)
        self.uiAdaptersSpinBox = QtGui.QSpinBox(self.tab_7)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uiAdaptersSpinBox.sizePolicy().hasHeightForWidth())
        self.uiAdaptersSpinBox.setSizePolicy(sizePolicy)
        self.uiAdaptersSpinBox.setMinimum(0)
        self.uiAdaptersSpinBox.setMaximum(32)
        self.uiAdaptersSpinBox.setObjectName(_fromUtf8("uiAdaptersSpinBox"))
        self.gridLayout_5.addWidget(self.uiAdaptersSpinBox, 0, 1, 1, 1)
        self.uiAdapterTypesLabel = QtGui.QLabel(self.tab_7)
        self.uiAdapterTypesLabel.setObjectName(_fromUtf8("uiAdapterTypesLabel"))
        self.gridLayout_5.addWidget(self.uiAdapterTypesLabel, 1, 0, 1, 1)
        self.uiAdapterTypesComboBox = QtGui.QComboBox(self.tab_7)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uiAdapterTypesComboBox.sizePolicy().hasHeightForWidth())
        self.uiAdapterTypesComboBox.setSizePolicy(sizePolicy)
        self.uiAdapterTypesComboBox.setObjectName(_fromUtf8("uiAdapterTypesComboBox"))
        self.gridLayout_5.addWidget(self.uiAdapterTypesComboBox, 1, 1, 1, 1)
        self.uiLegacyNetworkingCheckBox = QtGui.QCheckBox(self.tab_7)
        self.uiLegacyNetworkingCheckBox.setObjectName(_fromUtf8("uiLegacyNetworkingCheckBox"))
        self.gridLayout_5.addWidget(self.uiLegacyNetworkingCheckBox, 2, 0, 1, 2)
        spacerItem2 = QtGui.QSpacerItem(20, 261, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.gridLayout_5.addItem(spacerItem2, 3, 1, 1, 1)
        self.uiQemutabWidget.addTab(self.tab_7, _fromUtf8(""))
        self.tab_2 = QtGui.QWidget()
        self.tab_2.setObjectName(_fromUtf8("tab_2"))
        self.verticalLayout_2 = QtGui.QVBoxLayout(self.tab_2)
        self.verticalLayout_2.setObjectName(_fromUtf8("verticalLayout_2"))
        self.uiLinuxBootGroupBox = QtGui.QGroupBox(self.tab_2)
        self.uiLinuxBootGroupBox.setObjectName(_fromUtf8("uiLinuxBootGroupBox"))
        self.gridLayout_2 = QtGui.QGridLayout(self.uiLinuxBootGroupBox)
        self.gridLayout_2.setObjectName(_fromUtf8("gridLayout_2"))
        self.uiKernelImageLineEdit = QtGui.QLineEdit(self.uiLinuxBootGroupBox)
        self.uiKernelImageLineEdit.setObjectName(_fromUtf8("uiKernelImageLineEdit"))
        self.gridLayout_2.addWidget(self.uiKernelImageLineEdit, 1, 1, 1, 1)
        self.uiKernelCommandLineLabel = QtGui.QLabel(self.uiLinuxBootGroupBox)
        self.uiKernelCommandLineLabel.setObjectName(_fromUtf8("uiKernelCommandLineLabel"))
        self.gridLayout_2.addWidget(self.uiKernelCommandLineLabel, 2, 0, 1, 1)
        self.uiInitrdLabel = QtGui.QLabel(self.uiLinuxBootGroupBox)
        self.uiInitrdLabel.setObjectName(_fromUtf8("uiInitrdLabel"))
        self.gridLayout_2.addWidget(self.uiInitrdLabel, 0, 0, 1, 1)
        self.uiKernelImageLabel = QtGui.QLabel(self.uiLinuxBootGroupBox)
        self.uiKernelImageLabel.setObjectName(_fromUtf8("uiKernelImageLabel"))
        self.gridLayout_2.addWidget(self.uiKernelImageLabel, 1, 0, 1, 1)
        self.uiInitrdLineEdit = QtGui.QLineEdit(self.uiLinuxBootGroupBox)
        self.uiInitrdLineEdit.setObjectName(_fromUtf8("uiInitrdLineEdit"))
        self.gridLayout_2.addWidget(self.uiInitrdLineEdit, 0, 1, 1, 1)
        self.uiInitrdToolButton = QtGui.QToolButton(self.uiLinuxBootGroupBox)
        self.uiInitrdToolButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.uiInitrdToolButton.setObjectName(_fromUtf8("uiInitrdToolButton"))
        self.gridLayout_2.addWidget(self.uiInitrdToolButton, 0, 2, 1, 1)
        self.uiKernelImageToolButton = QtGui.QToolButton(self.uiLinuxBootGroupBox)
        self.uiKernelImageToolButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.uiKernelImageToolButton.setObjectName(_fromUtf8("uiKernelImageToolButton"))
        self.gridLayout_2.addWidget(self.uiKernelImageToolButton, 1, 2, 1, 1)
        self.uiKernelCommandLineEdit = QtGui.QLineEdit(self.uiLinuxBootGroupBox)
        self.uiKernelCommandLineEdit.setObjectName(_fromUtf8("uiKernelCommandLineEdit"))
        self.gridLayout_2.addWidget(self.uiKernelCommandLineEdit, 2, 1, 1, 2)
        self.verticalLayout_2.addWidget(self.uiLinuxBootGroupBox)
        self.uiOptimizationGroupBox = QtGui.QGroupBox(self.tab_2)
        self.uiOptimizationGroupBox.setObjectName(_fromUtf8("uiOptimizationGroupBox"))
        self.gridLayout = QtGui.QGridLayout(self.uiOptimizationGroupBox)
        self.gridLayout.setObjectName(_fromUtf8("gridLayout"))
        self.uiActivateCPUThrottlingCheckBox = QtGui.QCheckBox(self.uiOptimizationGroupBox)
        self.uiActivateCPUThrottlingCheckBox.setChecked(True)
        self.uiActivateCPUThrottlingCheckBox.setObjectName(_fromUtf8("uiActivateCPUThrottlingCheckBox"))
        self.gridLayout.addWidget(self.uiActivateCPUThrottlingCheckBox, 0, 0, 1, 2)
        self.uiCPUThrottlingLabel = QtGui.QLabel(self.uiOptimizationGroupBox)
        self.uiCPUThrottlingLabel.setObjectName(_fromUtf8("uiCPUThrottlingLabel"))
        self.gridLayout.addWidget(self.uiCPUThrottlingLabel, 1, 0, 1, 1)
        self.uiCPUThrottlingSpinBox = QtGui.QSpinBox(self.uiOptimizationGroupBox)
        self.uiCPUThrottlingSpinBox.setMinimum(1)
        self.uiCPUThrottlingSpinBox.setMaximum(800)
        self.uiCPUThrottlingSpinBox.setProperty("value", 100)
        self.uiCPUThrottlingSpinBox.setObjectName(_fromUtf8("uiCPUThrottlingSpinBox"))
        self.gridLayout.addWidget(self.uiCPUThrottlingSpinBox, 1, 1, 1, 1)
        self.uiProcessPriorityLabel = QtGui.QLabel(self.uiOptimizationGroupBox)
        self.uiProcessPriorityLabel.setObjectName(_fromUtf8("uiProcessPriorityLabel"))
        self.gridLayout.addWidget(self.uiProcessPriorityLabel, 2, 0, 1, 1)
        self.uiProcessPriorityComboBox = QtGui.QComboBox(self.uiOptimizationGroupBox)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uiProcessPriorityComboBox.sizePolicy().hasHeightForWidth())
        self.uiProcessPriorityComboBox.setSizePolicy(sizePolicy)
        self.uiProcessPriorityComboBox.setObjectName(_fromUtf8("uiProcessPriorityComboBox"))
        self.uiProcessPriorityComboBox.addItem(_fromUtf8(""))
        self.uiProcessPriorityComboBox.addItem(_fromUtf8(""))
        self.uiProcessPriorityComboBox.addItem(_fromUtf8(""))
        self.uiProcessPriorityComboBox.addItem(_fromUtf8(""))
        self.uiProcessPriorityComboBox.addItem(_fromUtf8(""))
        self.uiProcessPriorityComboBox.addItem(_fromUtf8(""))
        self.gridLayout.addWidget(self.uiProcessPriorityComboBox, 2, 1, 1, 1)
        self.verticalLayout_2.addWidget(self.uiOptimizationGroupBox)
        self.groupBox = QtGui.QGroupBox(self.tab_2)
        self.groupBox.setObjectName(_fromUtf8("groupBox"))
        self.gridLayout_3 = QtGui.QGridLayout(self.groupBox)
        self.gridLayout_3.setObjectName(_fromUtf8("gridLayout_3"))
        self.uiQemuOptionsLabel = QtGui.QLabel(self.groupBox)
        self.uiQemuOptionsLabel.setObjectName(_fromUtf8("uiQemuOptionsLabel"))
        self.gridLayout_3.addWidget(self.uiQemuOptionsLabel, 0, 0, 1, 1)
        self.uiQemuOptionsLineEdit = QtGui.QLineEdit(self.groupBox)
        self.uiQemuOptionsLineEdit.setObjectName(_fromUtf8("uiQemuOptionsLineEdit"))
        self.gridLayout_3.addWidget(self.uiQemuOptionsLineEdit, 0, 2, 1, 1)
        self.verticalLayout_2.addWidget(self.groupBox)
        spacerItem3 = QtGui.QSpacerItem(20, 90, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem3)
        self.uiQemutabWidget.addTab(self.tab_2, _fromUtf8(""))
        self.verticalLayout.addWidget(self.uiQemutabWidget)

        self.retranslateUi(QemuVMConfigPageWidget)
        self.uiQemutabWidget.setCurrentIndex(0)
        self.uiProcessPriorityComboBox.setCurrentIndex(3)
        QtCore.QMetaObject.connectSlotsByName(QemuVMConfigPageWidget)

    def retranslateUi(self, QemuVMConfigPageWidget):
        QemuVMConfigPageWidget.setWindowTitle(_translate("QemuVMConfigPageWidget", "QEMU VM configuration", None))
        self.uiQemuListLabel.setText(_translate("QemuVMConfigPageWidget", "Qemu binary:", None))
        self.uiConsolePortLabel.setText(_translate("QemuVMConfigPageWidget", "Console port:", None))
        self.uiNameLabel.setText(_translate("QemuVMConfigPageWidget", "VM name:", None))
        self.uiRamLabel.setText(_translate("QemuVMConfigPageWidget", "RAM:", None))
        self.uiRamSpinBox.setSuffix(_translate("QemuVMConfigPageWidget", " MB", None))
        self.uiQemutabWidget.setTabText(self.uiQemutabWidget.indexOf(self.tab), _translate("QemuVMConfigPageWidget", "General settings", None))
        self.uiHdaDiskImageLabel.setText(_translate("QemuVMConfigPageWidget", "Disk image (hda):", None))
        self.uiHdaDiskImageToolButton.setText(_translate("QemuVMConfigPageWidget", "&Browse...", None))
        self.uiHdbDiskImageLabel.setText(_translate("QemuVMConfigPageWidget", "Disk image (hdb):", None))
        self.uiHdbDiskImageToolButton.setText(_translate("QemuVMConfigPageWidget", "&Browse...", None))
        self.uiHdcDiskImageLabel.setText(_translate("QemuVMConfigPageWidget", "Disk image (hdc):", None))
        self.uiHdcDiskImageToolButton.setText(_translate("QemuVMConfigPageWidget", "&Browse...", None))
        self.uiHddDiskImageLabel.setText(_translate("QemuVMConfigPageWidget", "Disk image (hdd):", None))
        self.uiHddDiskImageToolButton.setText(_translate("QemuVMConfigPageWidget", "&Browse...", None))
        self.uiQemutabWidget.setTabText(self.uiQemutabWidget.indexOf(self.tab_3), _translate("QemuVMConfigPageWidget", "HDD", None))
        self.uiAdaptersLabel.setText(_translate("QemuVMConfigPageWidget", "Adapters:", None))
        self.uiAdapterTypesLabel.setText(_translate("QemuVMConfigPageWidget", "Type:", None))
        self.uiLegacyNetworkingCheckBox.setText(_translate("QemuVMConfigPageWidget", "Use the legacy networking mode", None))
        self.uiQemutabWidget.setTabText(self.uiQemutabWidget.indexOf(self.tab_7), _translate("QemuVMConfigPageWidget", "Network", None))
        self.uiLinuxBootGroupBox.setTitle(_translate("QemuVMConfigPageWidget", "Linux boot specific settings", None))
        self.uiKernelCommandLineLabel.setText(_translate("QemuVMConfigPageWidget", "Kernel command line:", None))
        self.uiInitrdLabel.setText(_translate("QemuVMConfigPageWidget", "Initial RAM disk (initrd):", None))
        self.uiKernelImageLabel.setText(_translate("QemuVMConfigPageWidget", "Kernel image:", None))
        self.uiInitrdToolButton.setText(_translate("QemuVMConfigPageWidget", "&Browse...", None))
        self.uiKernelImageToolButton.setText(_translate("QemuVMConfigPageWidget", "&Browse...", None))
        self.uiOptimizationGroupBox.setTitle(_translate("QemuVMConfigPageWidget", "Optimizations", None))
        self.uiActivateCPUThrottlingCheckBox.setText(_translate("QemuVMConfigPageWidget", "Activate CPU throttling", None))
        self.uiCPUThrottlingLabel.setText(_translate("QemuVMConfigPageWidget", "Percentage of CPU allowed:", None))
        self.uiCPUThrottlingSpinBox.setSuffix(_translate("QemuVMConfigPageWidget", " %", None))
        self.uiProcessPriorityLabel.setText(_translate("QemuVMConfigPageWidget", "Process priority:", None))
        self.uiProcessPriorityComboBox.setItemText(0, _translate("QemuVMConfigPageWidget", "Realtime", None))
        self.uiProcessPriorityComboBox.setItemText(1, _translate("QemuVMConfigPageWidget", "Very high", None))
        self.uiProcessPriorityComboBox.setItemText(2, _translate("QemuVMConfigPageWidget", "High", None))
        self.uiProcessPriorityComboBox.setItemText(3, _translate("QemuVMConfigPageWidget", "Normal", None))
        self.uiProcessPriorityComboBox.setItemText(4, _translate("QemuVMConfigPageWidget", "Low", None))
        self.uiProcessPriorityComboBox.setItemText(5, _translate("QemuVMConfigPageWidget", "Very low", None))
        self.groupBox.setTitle(_translate("QemuVMConfigPageWidget", "Aditional settings", None))
        self.uiQemuOptionsLabel.setText(_translate("QemuVMConfigPageWidget", "Options:", None))
        self.uiQemutabWidget.setTabText(self.uiQemutabWidget.indexOf(self.tab_2), _translate("QemuVMConfigPageWidget", "Advanced settings", None))




# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Configuration page for IOU device preferences.
"""

import copy
import os
import stat

from gns3.qt import QtCore, QtGui
from gns3.main_window import MainWindow
from gns3.dialogs.symbol_selection_dialog import SymbolSelectionDialog
from gns3.dialogs.configuration_dialog import ConfigurationDialog
from gns3.cloud.utils import UploadFilesThread
from gns3.utils.progress_dialog import ProgressDialog
from gns3.utils.file_copy_thread import FileCopyThread

from .. import IOU
from ..settings import IOU_DEVICE_SETTINGS
from ..ui.iou_device_preferences_page_ui import Ui_IOUDevicePreferencesPageWidget
from ..pages.iou_device_configuration_page import iouDeviceConfigurationPage
from ..dialogs.iou_device_wizard import IOUDeviceWizard


class IOUDevicePreferencesPage(QtGui.QWidget, Ui_IOUDevicePreferencesPageWidget):

    """
    QWidget preference page for IOU image & device preferences.
    """

    def __init__(self):
        QtGui.QWidget.__init__(self)
        self.setupUi(self)

        self._main_window = MainWindow.instance()
        self._iou_devices = {}
        self._items = []

        self.uiNewIOUDevicePushButton.clicked.connect(self._iouDeviceNewSlot)
        self.uiEditIOUDevicePushButton.clicked.connect(self._iouDeviceEditSlot)
        self.uiDeleteIOUDevicePushButton.clicked.connect(self._iouDeviceDeleteSlot)
        self.uiIOUDevicesTreeWidget.currentItemChanged.connect(self._iouDeviceChangedSlot)
        self.uiIOUDevicesTreeWidget.itemPressed.connect(self._iouDevicePressedSlot)

    def _iouDeviceChangedSlot(self, current, previous):
        """
        Loads a selected an IOU device from the tree widget.

        :param current: current QTreeWidgetItem instance
        :param previous: ignored
        """

        if not current:
            self.uiIOUDeviceInfoTreeWidget.clear()
            return

        self.uiEditIOUDevicePushButton.setEnabled(True)
        self.uiDeleteIOUDevicePushButton.setEnabled(True)
        key = current.data(0, QtCore.Qt.UserRole)
        iou_device = self._iou_devices[key]
        self._refreshInfo(iou_device)

    def _iouDeviceNewSlot(self):
        """
        Creates a new IOU device.
        """

        wizard = IOUDeviceWizard(self._iou_devices, parent=self)
        wizard.show()
        if wizard.exec_():

            new_device_settings = wizard.getSettings()
            key = "{server}:{name}".format(server=new_device_settings["server"], name=new_device_settings["name"])
            self._iou_devices[key] = IOU_DEVICE_SETTINGS.copy()
            self._iou_devices[key].update(new_device_settings)

            item = QtGui.QTreeWidgetItem(self.uiIOUDevicesTreeWidget)
            item.setText(0, self._iou_devices[key]["name"])
            item.setIcon(0, QtGui.QIcon(self._iou_devices[key]["default_symbol"]))
            item.setData(0, QtCore.Qt.UserRole, key)
            self._items.append(item)
            self.uiIOUDevicesTreeWidget.setCurrentItem(item)

            if new_device_settings["server"] == 'cloud':
                import logging
                log = logging.getLogger(__name__)

                # Start uploading the image to cloud files
                self._upload_image_progress_dialog = QtGui.QProgressDialog(
                    "Uploading image file {}".format(new_device_settings['image']), "Cancel", 0, 0, parent=self)
                self._upload_image_progress_dialog.setWindowModality(QtCore.Qt.WindowModal)
                self._upload_image_progress_dialog.setWindowTitle("IOU image upload")
                self._upload_image_progress_dialog.show()
                try:
                    src = self._iou_devices[key]['path']
                    # Eg: images/IOU/i86.bin
                    dst = 'images/IOU/{}'.format(self._iou_devices[key]['image'])
                    upload_thread = UploadFilesThread(self, MainWindow.instance().cloudSettings(), [(src, dst)])
                    upload_thread.completed.connect(self._imageUploadComplete)
                    upload_thread.start()
                except Exception as e:
                    self._upload_image_progress_dialog.reject()
                    log.error(e)
                    QtGui.QMessageBox.critical(self, "IOU image upload", "Error uploading IOU image: {}".format(e))

    def _imageUploadComplete(self):
        if self._upload_image_progress_dialog.wasCanceled():
            return
        self._upload_image_progress_dialog.accept()

    def _iouDeviceEditSlot(self):
        """
        Edits an IOU device.
        """

        item = self.uiIOUDevicesTreeWidget.currentItem()
        if item:
            key = item.data(0, QtCore.Qt.UserRole)
            iou_device = self._iou_devices[key]
            dialog = ConfigurationDialog(iou_device["name"], iou_device, iouDeviceConfigurationPage(), parent=self)
            dialog.show()
            if dialog.exec_():
                if iou_device["name"] != item.text(0):
                    new_key = "{server}:{name}".format(server=iou_device["server"], name=iou_device["name"])
                    if new_key in self._iou_devices:
                        QtGui.QMessageBox.critical(self, "IOU device", "IOU device name {} already exists for server {}".format(iou_device["name"],
                                                                                                                                iou_device["server"]))
                        iou_device["name"] = item.text(0)
                        return
                    self._iou_devices[new_key] = self._iou_devices[key]
                    del self._iou_devices[key]
                    item.setText(0, iou_device["name"])
                    item.setData(0, QtCore.Qt.UserRole, new_key)
                self._refreshInfo(iou_device)

    def _iouDeviceDeleteSlot(self):
        """
        Deletes an IOU device.
        """

        item = self.uiIOUDevicesTreeWidget.currentItem()
        if item:
            key = item.data(0, QtCore.Qt.UserRole)
            del self._iou_devices[key]
            self.uiIOUDevicesTreeWidget.takeTopLevelItem(self.uiIOUDevicesTreeWidget.indexOfTopLevelItem(item))

    def _createSectionItem(self, name):

        section_item = QtGui.QTreeWidgetItem(self.uiIOUDeviceInfoTreeWidget)
        section_item.setText(0, name)
        font = section_item.font(0)
        font.setBold(True)
        section_item.setFont(0, font)
        return section_item

    def _refreshInfo(self, iou_device):

        self.uiIOUDeviceInfoTreeWidget.clear()

        # fill out the General section
        section_item = self._createSectionItem("General")
        QtGui.QTreeWidgetItem(section_item, ["Name:", iou_device["name"]])
        QtGui.QTreeWidgetItem(section_item, ["Server:", iou_device["server"]])
        QtGui.QTreeWidgetItem(section_item, ["Image:", iou_device["image"]])
        QtGui.QTreeWidgetItem(section_item, ["Initial config:", iou_device["initial_config"]])

        if iou_device["use_default_iou_values"]:
            QtGui.QTreeWidgetItem(section_item, ["RAM:", "default"])
            QtGui.QTreeWidgetItem(section_item, ["NVRAM:", "default"])
        else:
            QtGui.QTreeWidgetItem(section_item, ["RAM:", "{} MiB".format(iou_device["ram"])])
            QtGui.QTreeWidgetItem(section_item, ["NVRAM:", "{} KiB".format(iou_device["nvram"])])

        # fill out the Network section
        section_item = self._createSectionItem("Network")
        QtGui.QTreeWidgetItem(section_item, ["Ethernet adapters:", "{} ({} interfaces)".format(iou_device["ethernet_adapters"],
                                                                                               iou_device["ethernet_adapters"] * 4)])
        QtGui.QTreeWidgetItem(section_item, ["Serial adapters:", "{} ({} interfaces)".format(iou_device["serial_adapters"],
                                                                                             iou_device["serial_adapters"] * 4)])

        self.uiIOUDeviceInfoTreeWidget.expandAll()
        self.uiIOUDeviceInfoTreeWidget.resizeColumnToContents(0)
        self.uiIOUDeviceInfoTreeWidget.resizeColumnToContents(1)

    @staticmethod
    def getIOUImage(parent):
        """

        :param parent: parent widget

        :return: path to the IOU image or None
        """

        destination_directory = os.path.join(MainWindow.instance().imagesDirPath(), "IOU")
        path, _ = QtGui.QFileDialog.getOpenFileNameAndFilter(parent,
                                                             "Select an IOU image",
                                                             destination_directory,
                                                             "All files (*)",
                                                             "IOU image (*.bin *.image)")

        if not path:
            return

        if not os.access(path, os.R_OK):
            QtGui.QMessageBox.critical(parent, "IOU image", "Cannot read {}".format(path))
            return

        try:
            with open(path, "rb") as f:
                # read the first 7 bytes of the file.
                elf_header_start = f.read(7)
        except OSError as e:
            QtGui.QMessageBox.critical(parent, "IOU image", "Cannot read ELF magic number: {}".format(e))
            return

        # file must start with the ELF magic number, be 32-bit, little endian and have an ELF version of 1
        # normal IOS image are big endian!
        if elf_header_start != b'\x7fELF\x01\x01\x01':
            QtGui.QMessageBox.critical(parent, "IOU image", "Sorry, this is not a valid IOU image!")
            return

        try:
            os.makedirs(destination_directory)
        except FileExistsError:
            pass
        except OSError as e:
            QtGui.QMessageBox.critical(parent, "IOU images directory", "Could not create the IOU images directory {}: {}".format(destination_directory, e))
            return

        if os.path.normpath(os.path.dirname(path)) != destination_directory:
            # the IOU image is not in the default images directory
            reply = QtGui.QMessageBox.question(parent,
                                               "IOU image",
                                               "Would you like to copy {} to the default images directory".format(os.path.basename(path)),
                                               QtGui.QMessageBox.Yes,
                                               QtGui.QMessageBox.No)
            if reply == QtGui.QMessageBox.Yes:
                destination_path = os.path.join(destination_directory, os.path.basename(path))
                thread = FileCopyThread(path, destination_path)
                progress_dialog = ProgressDialog(thread, "Project", "Copying {}".format(os.path.basename(path)), "Cancel", busy=True, parent=parent)
                thread.deleteLater()
                progress_dialog.show()
                progress_dialog.exec_()
                errors = progress_dialog.errors()
                if errors:
                    QtGui.QMessageBox.critical(parent, "IOS image", "{}".format("".join(errors)))
                else:
                    path = destination_path
                    mode = os.stat(path).st_mode
                    os.chmod(path, mode | stat.S_IXUSR)

        if not os.access(path, os.X_OK):
            QtGui.QMessageBox.warning(parent, "IOU image", "{} is not executable".format(path))

        return path

    def _iouDevicePressedSlot(self, item, column):
        """
        Slot for item pressed.

        :param item: ignored
        :param column: ignored
        """

        if QtGui.QApplication.mouseButtons() & QtCore.Qt.RightButton:
            self._showContextualMenu()

    def _showContextualMenu(self):
        """
        Contextual menu.
        """

        menu = QtGui.QMenu()
        change_symbol_action = QtGui.QAction("Change symbol", menu)
        change_symbol_action.setIcon(QtGui.QIcon(":/icons/node_conception.svg"))
        self.connect(change_symbol_action, QtCore.SIGNAL('triggered()'), self._changeSymbolSlot)
        menu.addAction(change_symbol_action)
        menu.exec_(QtGui.QCursor.pos())

    def _changeSymbolSlot(self):
        """
        Change a symbol for an IOU device.
        """

        item = self.uiIOUDevicesTreeWidget.currentItem()
        if item:
            key = item.data(0, QtCore.Qt.UserRole)
            iou_device = self._iou_devices[key]
            dialog = SymbolSelectionDialog(self, symbol=iou_device["default_symbol"], category=iou_device["category"])
            dialog.show()
            if dialog.exec_():
                normal_symbol, selected_symbol = dialog.getSymbols()
                category = dialog.getCategory()
                item.setIcon(0, QtGui.QIcon(normal_symbol))
                iou_device["default_symbol"] = normal_symbol
                iou_device["hover_symbol"] = selected_symbol
                iou_device["category"] = category

    def loadPreferences(self):
        """
        Loads the IOU devices preferences.
        """

        iou_module = IOU.instance()
        self._iou_devices = copy.deepcopy(iou_module.iouDevices())
        self._items.clear()

        for key, iou_device in self._iou_devices.items():
            item = QtGui.QTreeWidgetItem(self.uiIOUDevicesTreeWidget)
            item.setText(0, iou_device["name"])
            item.setIcon(0, QtGui.QIcon(iou_device["default_symbol"]))
            item.setData(0, QtCore.Qt.UserRole, key)
            self._items.append(item)

        if self._items:
            self.uiIOUDevicesTreeWidget.setCurrentItem(self._items[0])
            self.uiIOUDevicesTreeWidget.sortByColumn(0, QtCore.Qt.AscendingOrder)

    def savePreferences(self):
        """
        Saves the IOU devices preferences.
        """

        # self._iouImageSaveSlot()
        IOU.instance().setIOUDevices(self._iou_devices)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
import tempfile
import pkg_resources

try:
    pkg_resources.set_extraction_path(tempfile.mkdtemp())
except ValueError:
    pass  # If the path is already set the module throw an error


def get_resource(resource_name):

    resource_path = None
    if hasattr(sys, "frozen") and sys.platform.startswith("darwin"):
        resource_name = os.path.join("../Resources", resource_name)
    if hasattr(sys, "frozen") and os.path.exists(resource_name):
        resource_path = os.path.normpath(os.path.join(os.getcwd(), resource_name))
    elif not hasattr(sys, "frozen") and pkg_resources.resource_exists("gns3", resource_name):
        resource_path = pkg_resources.resource_filename("gns3", resource_name)
        resource_path = os.path.normpath(resource_path)
    return resource_path



# -*- coding: utf-8 -*-
#
# Copyright (C) 2014 GNS3 Technologies Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Graphical view on the scene where items are drawn.
"""

import logging
import os
import pickle
import functools

from .qt import QtCore, QtGui, QtNetwork
from .servers import Servers
from .items.node_item import NodeItem
from .dialogs.node_configurator_dialog import NodeConfiguratorDialog
from .link import Link
from .node import Node
from .modules import MODULES
from .modules.builtin.cloud import Cloud
from .modules.module_error import ModuleError
from .settings import GRAPHICS_VIEW_SETTINGS, GRAPHICS_VIEW_SETTING_TYPES
from .topology import Topology
from .ports.port import Port
from .dialogs.style_editor_dialog import StyleEditorDialog
from .dialogs.text_editor_dialog import TextEditorDialog
from .dialogs.symbol_selection_dialog import SymbolSelectionDialog
from .dialogs.idlepc_dialog import IdlePCDialog
from .local_config import LocalConfig

# link items
from .items.link_item import LinkItem
from .items.ethernet_link_item import EthernetLinkItem
from .items.serial_link_item import SerialLinkItem

# other items
from .items.note_item import NoteItem
from .items.shape_item import ShapeItem
from .items.rectangle_item import RectangleItem
from .items.ellipse_item import EllipseItem
from .items.image_item import ImageItem

log = logging.getLogger(__name__)


class GraphicsView(QtGui.QGraphicsView):

    """
    Graphics view that displays the scene.

    :param parent: parent widget
    """

    def __init__(self, parent):

        # Our parent is the central widget which parent
        # is the main window.
        self._main_window = parent.parent()

        QtGui.QGraphicsView.__init__(self, parent)
        self._settings = {}
        self._loadSettings()

        self._adding_link = False
        self._adding_note = False
        self._adding_rectangle = False
        self._adding_ellipse = False
        self._newlink = None
        self._dragging = False
        self._last_mouse_position = None
        self._topology = Topology.instance()

        # set the scene
        scene = QtGui.QGraphicsScene(parent=self)
        width = self._settings["scene_width"]
        height = self._settings["scene_height"]
        scene.setSceneRect(-(width / 2), -(height / 2), width, height)
        self.setScene(scene)

        # set the custom flags for this view
        self.setDragMode(QtGui.QGraphicsView.RubberBandDrag)
        self.setCacheMode(QtGui.QGraphicsView.CacheBackground)
        self.setRenderHint(QtGui.QPainter.Antialiasing)
        self.setTransformationAnchor(QtGui.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QtGui.QGraphicsView.AnchorViewCenter)

        self._local_addresses = ['0.0.0.0', '127.0.0.1', 'localhost', '::1', '0:0:0:0:0:0:0:1', '::', QtNetwork.QHostInfo.localHostName()]

    def reset(self):
        """
        Remove all the items from the scene and
        reset the instances count.
        """

        # reset the modules
        for module in MODULES:
            instance = module.instance()
            instance.reset()

        # reset instance IDs for
        # nodes, links and ports
        Node.reset()
        Link.reset()
        Port.reset()

        # reset the topology
        self._topology.reset()

        # clear the topology summary
        self._main_window.uiTopologySummaryTreeWidget.clear()

        # clear all objects on the scene
        self.scene().clear()

    def _loadSettings(self):
        """
        Loads the settings from the persistent settings file.
        """

        local_config = LocalConfig.instance()
        # restore the graphics view settings from QSettings (for backward compatibility)
        legacy_settings = {}
        settings = QtCore.QSettings()
        settings.beginGroup(self.__class__.__name__)
        for name in GRAPHICS_VIEW_SETTINGS.keys():
            if settings.contains(name):
                legacy_settings[name] = settings.value(name, type=GRAPHICS_VIEW_SETTING_TYPES[name])
        settings.remove("")
        settings.endGroup()

        if legacy_settings:
            local_config.saveSectionSettings(self.__class__.__name__, legacy_settings)

        self._settings = local_config.loadSectionSettings(self.__class__.__name__, GRAPHICS_VIEW_SETTINGS)

    def settings(self):
        """
        Returns the graphics view settings.

        :returns: settings dictionary
        """

        return self._settings

    def setSettings(self, new_settings):
        """
        Set new graphics view settings.

        :param new_settings: settings dictionary
        """

        # save the settings
        self._settings.update(new_settings)
        LocalConfig.instance().saveSectionSettings(self.__class__.__name__, self._settings)

    def addingLinkSlot(self, enabled):
        """
        Slot to receive events from MainWindow
        when a user has clicked on "Add a link" button.

        :param enable: either the user is adding a link or not (boolean)
        """

        if enabled:
            self.setCursor(QtCore.Qt.CrossCursor)
        else:
            if self._newlink:
                self.scene().removeItem(self._newlink)
                self._newlink = None
            self.setCursor(QtCore.Qt.ArrowCursor)
        self._adding_link = enabled

    def addNote(self, state):
        """
        Adds a note.

        :param state: boolean
        """

        if state:
            self._adding_note = True
            self.setCursor(QtCore.Qt.IBeamCursor)
        else:
            self._adding_note = False
            self.setCursor(QtCore.Qt.ArrowCursor)

    def addRectangle(self, state):
        """
        Adds a rectangle.

        :param state: boolean
        """

        if state:
            self._adding_rectangle = True
            self.setCursor(QtCore.Qt.PointingHandCursor)
        else:
            self._adding_rectangle = False
            self.setCursor(QtCore.Qt.ArrowCursor)

    def addEllipse(self, state):
        """
        Adds an ellipse.

        :param state: boolean
        """

        if state:
            self._adding_ellipse = True
            self.setCursor(QtCore.Qt.PointingHandCursor)
        else:
            self._adding_ellipse = False
            self.setCursor(QtCore.Qt.ArrowCursor)

    def addImage(self, pixmap, image_path):
        """
        Adds an image.

        :param pixmap: QPixmap instance
        :param image_path: path to the image
        """

        image_item = ImageItem(pixmap, image_path)
        # center the image on the scene
        x = image_item.pos().x() - (image_item.boundingRect().width() / 2)
        y = image_item.pos().y() - (image_item.boundingRect().height() / 2)
        image_item.setPos(x, y)
        self.scene().addItem(image_item)
        self._topology.addImage(image_item)

    def addLink(self, source_node, source_port, destination_node, destination_port):
        """
        Creates a Link instance representing a connection between 2 devices.

        :param source_node: source Node instance
        :param source_port: source Port instance
        :param destination_node: destination Node instance
        :param destination_port: destination Port instance
        """

        link = Link(source_node, source_port, destination_node, destination_port)

        # connect the signals that let the graphics view knows about events such as
        # a new link creation or deletion.
        link.add_link_signal.connect(self.addLinkSlot)
        link.delete_link_signal.connect(self.deleteLinkSlot)
        self._topology.addLink(link)

    def addLinkSlot(self, link_id):
        """
        Slot to receive events from Link instances
        when a link has been created.

        :param link_id: link identifier
        """

        link = self._topology.getLink(link_id)
        source_item = None
        destination_item = None
        source_port = link.sourcePort()
        destination_port = link.destinationPort()

        # find the correct source and destination node items
        for item in self.scene().items():
            if isinstance(item, NodeItem):
                if item.node().id() == link.sourceNode().id():
                    source_item = item
                if item.node().id() == link.destinationNode().id():
                    destination_item = item
            if source_item and destination_item:
                break

        if not source_item or not destination_item:
            print("Could not find a source or destination item for the link!")
            self.deleteLinkSlot(link_id)
            return

        # ugly multi-link management
        # FIXME: taken from old GNS3 and has a bug!
        multi = 0
        d1 = 0
        d2 = 1
        link_items = source_item.links()
        for link_item in link_items:
            if link_item.destinationItem().node().id() == destination_item.node().id():
                d1 += 1
            if link_item.sourceItem().node().id() == destination_item.node().id():
                d2 += 1

        if len(link_items) > 0:
            if d2 - d1 == 2:
                source_port, destination_port = destination_port, source_port
                source_item, destination_item = destination_item, source_item
                multi = d1 + 1
            elif d1 >= d2:
                source_port, destination_port = destination_port, source_port
                source_item, destination_item = destination_item, source_item
                multi = d2
            else:
                multi = d1

        # MAX 7 links on the scene between 2 nodes
        if multi > 3:
            multi = 0

        if source_item == destination_item:
            multi = 0

        if link.sourcePort().linkType() == "Serial" or (source_port.isStub() and link.destinationPort().linkType() == "Serial"):
            link_item = SerialLinkItem(source_item, source_port, destination_item, destination_port, link, multilink=multi)
        else:
            link_item = EthernetLinkItem(source_item, source_port, destination_item, destination_port, link, multilink=multi)
        self.scene().addItem(link_item)

    def deleteLinkSlot(self, link_id):
        """
        Slot to receive events from Link instances
        when a link has been deleted.

        :param link_id: link identifier
        """

        link = self._topology.getLink(link_id)
        # disconnect the signals just in case...
        link.add_link_signal.disconnect()
        link.delete_link_signal.disconnect()
        self._topology.removeLink(link)

    def _userNodeLinking(self, event, item):
        """
        Handles node linking by the user.

        :param event: QMouseEvent instance
        :param item: NodeItem instance
        """

        # link addition code
        if not self._newlink:
            source_item = item
            source_port = source_item.connectToPort()
            if not source_port:
                return
            if not source_item.node().initialized():
                QtGui.QMessageBox.critical(self, "Connection", "This node hasn't been initialized correctly")
                return
            if not source_port.isFree():
                QtGui.QMessageBox.critical(self, "Connection", "Port {} isn't free".format(source_port.name()))
                return
            if source_port.linkType() == "Serial":
                self._newlink = SerialLinkItem(source_item, source_port, self.mapToScene(event.pos()), None, adding_flag=True)
            else:
                self._newlink = EthernetLinkItem(source_item, source_port, self.mapToScene(event.pos()), None, adding_flag=True)
            self.scene().addItem(self._newlink)
        else:
            source_item = self._newlink.sourceItem()
            source_port = self._newlink.sourcePort()
            destination_item = item
            if source_item == destination_item:
                QtGui.QMessageBox.critical(self, "Connection", "Cannot connect to itself!")
                return
            destination_port = destination_item.connectToPort()
            if not destination_port:
                return
            if not destination_item.node().initialized():
                QtGui.QMessageBox.critical(self, "Connection", "This node hasn't been initialized correctly")
                return
            if not destination_port.isFree():
                QtGui.QMessageBox.critical(self, "Connection", "Port {} isn't free".format(destination_port.name()))
                return
            if source_port.isStub() or destination_port.isStub():
                pass
            elif source_port.linkType() != destination_port.linkType():
                QtGui.QMessageBox.critical(self, "Connection", "Cannot connect this port!")
                return

            if isinstance(source_item.node(), Cloud) and isinstance(destination_item.node(), Cloud):
                QtGui.QMessageBox.critical(self, "Connection", "Sorry, you cannot connect a cloud to another cloud!")
                return

            source_host = source_item.node().server().host
            destination_host = destination_item.node().server().host

            # check that the node can be connected to a cloud
            if (isinstance(source_item.node(), Cloud) or isinstance(destination_item.node(), Cloud)) and source_host != destination_host:
                QtGui.QMessageBox.critical(self, "Connection", "This device can only be connected to a cloud on the same host")
                return

            # check if the 2 nodes can communicate
            if (source_host in self._local_addresses and destination_host not in self._local_addresses) or \
               (destination_host in self._local_addresses and source_host not in self._local_addresses):
                QtGui.QMessageBox.critical(self, "Connection", "Server {} cannot communicate with server {}, most likely because your local server host binding is set to a local address".format(source_host, destination_host))
                return

            self.scene().removeItem(self._newlink)
            self.addLink(source_item.node(), source_port, destination_item.node(), destination_port)
            self._newlink = None

    def mousePressEvent(self, event):
        """
        Handles all mouse press events.

        :param event: QMouseEvent instance
        """

        is_not_link = True
        item = self.itemAt(event.pos())
        if item and isinstance(item, LinkItem):
            is_not_link = False
        else:
            for it in self.scene().items():
                if isinstance(it, LinkItem):
                    it.setHovered(False)

        if (event.buttons() == QtCore.Qt.LeftButton and event.modifiers() == QtCore.Qt.ShiftModifier) or event.buttons() == QtCore.Qt.MidButton:
            # checks to see if either the middle mouse is pressed
            # or a combination of left mouse button and SHIT key are pressed to start dragging the view
            self._last_mouse_position = self.mapFromGlobal(event.globalPos())
            self._dragging = True
            self.setCursor(QtCore.Qt.ClosedHandCursor)
            return

        if is_not_link and item and event.modifiers() == QtCore.Qt.ControlModifier and event.button() == QtCore.Qt.LeftButton and item and not self._adding_link:
            # manual selection using CTRL
            if item.isSelected():
                item.setSelected(False)
            else:
                item.setSelected(True)
        elif is_not_link and event.button() == QtCore.Qt.RightButton and not self._adding_link:
            if item:
                # Prevent right clicking on a selected item from de-selecting all other items
                if not item.isSelected():
                    if not event.modifiers() & QtCore.Qt.ControlModifier:
                        for it in self.scene().items():
                            it.setSelected(False)
                    if item.zValue() < 0:
                        item.setFlag(item.ItemIsSelectable, True)
                    item.setSelected(True)
                    self._showDeviceContextualMenu(QtGui.QCursor.pos())
                    if item.zValue() < 0:
                        item.setFlag(item.ItemIsSelectable, False)

                else:
                    self._showDeviceContextualMenu(QtGui.QCursor.pos())
            # when more than one item is selected display the contextual menu even if mouse is not above an item
            elif len(self.scene().selectedItems()) > 1:
                self._showDeviceContextualMenu(QtGui.QCursor.pos())
        elif item and isinstance(item, NodeItem) and self._adding_link and event.button() == QtCore.Qt.LeftButton:
            self._userNodeLinking(event, item)
        elif event.button() == QtCore.Qt.LeftButton and self._adding_note:
            note = NoteItem()
            note.setPos(self.mapToScene(event.pos()))
            pos_x = note.pos().x()
            pos_y = note.pos().y() - (note.boundingRect().height() / 2)
            note.setPos(pos_x, pos_y)
            self.scene().addItem(note)
            self._topology.addNote(note)
            note.editText()
            self._main_window.uiAddNoteAction.setChecked(False)
            self.setCursor(QtCore.Qt.ArrowCursor)
            self._adding_note = False
        elif event.button() == QtCore.Qt.LeftButton and self._adding_rectangle:
            rectangle = RectangleItem(self.mapToScene(event.pos()))
            self.scene().addItem(rectangle)
            self._topology.addRectangle(rectangle)
            self._main_window.uiDrawRectangleAction.setChecked(False)
            self.setCursor(QtCore.Qt.ArrowCursor)
            self._adding_rectangle = False
        elif event.button() == QtCore.Qt.LeftButton and self._adding_ellipse:
            ellipse = EllipseItem(self.mapToScene(event.pos()))
            self.scene().addItem(ellipse)
            self._topology.addEllipse(ellipse)
            self._main_window.uiDrawEllipseAction.setChecked(False)
            self.setCursor(QtCore.Qt.ArrowCursor)
            self._adding_ellipse = False
        else:
            QtGui.QGraphicsView.mousePressEvent(self, event)

    def mouseReleaseEvent(self, event):
        """
        Handles all mouse release events.

        :param: QMouseEvent instance
        """

        item = self.itemAt(event.pos())
        # If the left  mouse button is not still pressed TOGETHER with the SHIFT key and neither is the middle button
        # this means the user is no longer trying to drag the view
        if self._dragging and not (event.buttons() == QtCore.Qt.LeftButton and event.modifiers() == QtCore.Qt.ShiftModifier) and not event.buttons() & QtCore.Qt.MidButton:
            self._dragging = False
            self.setCursor(QtCore.Qt.ArrowCursor)
        else:
            if item is not None and not event.modifiers() & QtCore.Qt.ControlModifier:
                item.setSelected(True)
            QtGui.QGraphicsView.mouseReleaseEvent(self, event)

    def wheelEvent(self, event):
        """
        Handles zoom in or out using the mouse wheel.

        :param: QWheelEvent instance
        """

        if event.modifiers() == QtCore.Qt.ControlModifier and event.orientation() == QtCore.Qt.Vertical:
            # CTRL is pressed then use the mouse wheel to zoom in or out.
            self.scaleView(pow(2.0, event.delta() / 240.0))
        else:
            QtGui.QGraphicsView.wheelEvent(self, event)

    def scaleView(self, scale_factor):
        """
        Scales the view (zoom in and out).
        """

        factor = self.matrix().scale(scale_factor, scale_factor).mapRect(QtCore.QRectF(0, 0, 1, 1)).width()
        if (factor < 0.10 or factor > 10):
            return
        self.scale(scale_factor, scale_factor)

    def keyPressEvent(self, event):
        """
        Handles all key press events for this view.

        :param event: QKeyEvent
        """

        if event.key() == QtCore.Qt.Key_Delete:
            # check if we are editing an NoteItem instance, then send the delete key event to it
            for item in self.scene().selectedItems():
                if isinstance(item, NoteItem) and item.hasFocus():
                    QtGui.QGraphicsView.keyPressEvent(self, event)
                    return
            self.deleteActionSlot()
            QtGui.QGraphicsView.keyPressEvent(self,event)
        else:
            QtGui.QGraphicsView.keyPressEvent(self, event)

    def mouseMoveEvent(self, event):
        """
        Handles all mouse move events (mouse tracking has been enabled).

        :param event: QMouseEvent instance
        """

        if self._dragging:
            # This if statement event checks to see if the user is dragging the scene
            # if so it sets the value of the scene scroll bars based on the change between
            # the previous and current mouse position
            mapped_global_pos = self.mapFromGlobal(event.globalPos())
            hBar = self.horizontalScrollBar()
            vBar = self.verticalScrollBar()
            delta = mapped_global_pos - self._last_mouse_position
            hBar.setValue(hBar.value() + (delta.x() if QtGui.QApplication.isRightToLeft() else -delta.x()))
            vBar.setValue(vBar.value() - delta.y())
            self._last_mouse_position = mapped_global_pos
        if self._adding_link and self._newlink:
            # update the mouse position when the user is adding a link.
            self._newlink.setMousePoint(self.mapToScene(event.pos()))
            event.ignore()
        else:
            item = self.itemAt(event.pos())
            if item:
                # show item coords in the status bar
                coords = "X: {} Y: {} Z: {}".format(item.x(), item.y(), item.zValue())
                self._main_window.uiStatusBar.showMessage(coords, 2000)
            QtGui.QGraphicsView.mouseMoveEvent(self, event)

    def mouseDoubleClickEvent(self, event):
        """
        Handles all mouse double click events.

        :param event: QMouseEvent instance
        """

        item = self.itemAt(event.pos())
        if not self._adding_link and isinstance(item, NodeItem) and item.node().initialized():
            item.setSelected(True)
            if isinstance(item, NodeItem):
                self.consoleToNode(item.node())
            else:
                self.configureSlot()
        else:
            QtGui.QGraphicsView.mouseDoubleClickEvent(self, event)

    def configureSlot(self, items=None):
        """
        Opens the node configurator.
        """

        if not items:
            items = self.scene().selectedItems()
        node_configurator = NodeConfiguratorDialog(items, self._main_window)
        node_configurator.setModal(True)
        node_configurator.show()
        node_configurator.exec_()
        for item in items:
            item.setSelected(False)

    def dragMoveEvent(self, event):
        """
        Handles all drag move events.

        :param event: QDragMoveEvent instance
        """

        # check if what is dragged is handled by this view
        if event.mimeData().hasFormat("application/x-gns3-node") or event.mimeData().hasFormat("text/uri-list"):
            event.acceptProposedAction()
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Handles all drop events.

        :param event: QDropEvent instance
        """

        # check if what has been dropped is handled by this view
        if event.mimeData().hasFormat("application/x-gns3-node"):
            data = event.mimeData().data("application/x-gns3-node")
            # load the pickled node data
            node_data = pickle.loads(data)
            event.setDropAction(QtCore.Qt.CopyAction)
            event.accept()
            if event.keyboardModifiers() == QtCore.Qt.ShiftModifier:
                max_nodes_per_line = 10  # max number of nodes on a single line
                offset = 100  # spacing between elements
                integer, ok = QtGui.QInputDialog.getInteger(self, "Nodes", "Number of nodes:", 2, 1, 100, 1)
                if ok:
                    for node_number in range(integer):
                        node_item = self.createNode(node_data, event.pos())
                        if node_item is None:
                            # stop if there is any error
                            break
                        x = node_item.pos().x() - (node_item.boundingRect().width() / 2) + (node_number % max_nodes_per_line) * offset
                        y = node_item.pos().y() - (node_item.boundingRect().height() / 2) + (node_number // max_nodes_per_line) * offset
                        node_item.setPos(x, y)
            else:
                self.createNode(node_data, event.pos())
        elif event.mimeData().hasFormat("text/uri-list") and event.mimeData().hasUrls():
            if len(event.mimeData().urls()) > 1:
                QtGui.QMessageBox.critical(self, "Project files", "Please drop only one file")
                return
            path = event.mimeData().urls()[0].toLocalFile()
            if os.path.isfile(path) and self._main_window.checkForUnsavedChanges():
                self._main_window.loadProject(path)
            event.acceptProposedAction()
        else:
            event.ignore()

    def _showDeviceContextualMenu(self, pos):
        """
        Create and display the device contextual menu on the view.

        :param pos: position where to display the menu
        """

        menu = QtGui.QMenu()
        self.populateDeviceContextualMenu(menu)
        menu.exec_(pos)
        menu.clear()

    def populateDeviceContextualMenu(self, menu):
        """
        Adds device actions to the device contextual menu.

        :param menu: QMenu instance
        """

        items = self.scene().selectedItems()
        if not items:
            return

        if True in list(map(lambda item: isinstance(item, NodeItem), items)):
            configure_action = QtGui.QAction("Configure", menu)
            configure_action.setIcon(QtGui.QIcon(':/icons/configuration.svg'))
            configure_action.triggered.connect(self.configureActionSlot)
            menu.addAction(configure_action)

            # Action: Change hostname
            change_hostname_action = QtGui.QAction("Change hostname", menu)
            change_hostname_action.setIcon(QtGui.QIcon(':/icons/show-hostname.svg'))
            self.connect(change_hostname_action, QtCore.SIGNAL('triggered()'), self.changeHostnameActionSlot)
            menu.addAction(change_hostname_action)

            # Action: Change symbol
            change_symbol_action = QtGui.QAction("Change symbol", menu)
            change_symbol_action.setIcon(QtGui.QIcon(':/icons/node_conception.svg'))
            self.connect(change_symbol_action, QtCore.SIGNAL('triggered()'), self.changeSymbolActionSlot)
            menu.addAction(change_symbol_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "console"), items)):
            console_action = QtGui.QAction("Console", menu)
            console_action.setIcon(QtGui.QIcon(':/icons/console.svg'))
            console_action.triggered.connect(self.consoleActionSlot)
            menu.addAction(console_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "auxConsole"), items)):
            aux_console_action = QtGui.QAction("Auxiliary console", menu)
            aux_console_action.setIcon(QtGui.QIcon(':/icons/aux-console.svg'))
            aux_console_action.triggered.connect(self.auxConsoleActionSlot)
            menu.addAction(aux_console_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "importConfig"), items)):
            import_config_action = QtGui.QAction("Import config", menu)
            import_config_action.setIcon(QtGui.QIcon(':/icons/import_config.svg'))
            import_config_action.triggered.connect(self.importConfigActionSlot)
            menu.addAction(import_config_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "exportConfig"), items)):
            export_config_action = QtGui.QAction("Export config", menu)
            export_config_action.setIcon(QtGui.QIcon(':/icons/export_config.svg'))
            export_config_action.triggered.connect(self.exportConfigActionSlot)
            menu.addAction(export_config_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "saveConfig"), items)):
            save_config_action = QtGui.QAction("Save config", menu)
            save_config_action.setIcon(QtGui.QIcon(':/icons/save.svg'))
            save_config_action.triggered.connect(self.saveConfigActionSlot)
            menu.addAction(save_config_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "startPacketCapture"), items)):
            capture_action = QtGui.QAction("Capture", menu)
            capture_action.setIcon(QtGui.QIcon(':/icons/inspect.svg'))
            capture_action.triggered.connect(self.captureActionSlot)
            menu.addAction(capture_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "idlepc"), items)):
            idlepc_action = QtGui.QAction("Idle-PC", menu)
            idlepc_action.setIcon(QtGui.QIcon(':/icons/calculate.svg'))
            idlepc_action.triggered.connect(self.idlepcActionSlot)
            menu.addAction(idlepc_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "start"), items)):
            start_action = QtGui.QAction("Start", menu)
            start_action.setIcon(QtGui.QIcon(':/icons/start.svg'))
            start_action.triggered.connect(self.startActionSlot)
            menu.addAction(start_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "suspend"), items)):
            suspend_action = QtGui.QAction("Suspend", menu)
            suspend_action.setIcon(QtGui.QIcon(':/icons/pause.svg'))
            suspend_action.triggered.connect(self.suspendActionSlot)
            menu.addAction(suspend_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "stop"), items)):
            stop_action = QtGui.QAction("Stop", menu)
            stop_action.setIcon(QtGui.QIcon(':/icons/stop.svg'))
            stop_action.triggered.connect(self.stopActionSlot)
            menu.addAction(stop_action)

        if True in list(map(lambda item: isinstance(item, NodeItem) and hasattr(item.node(), "reload"), items)):
            reload_action = QtGui.QAction("Reload", menu)
            reload_action.setIcon(QtGui.QIcon(':/icons/reload.svg'))
            reload_action.triggered.connect(self.reloadActionSlot)
            menu.addAction(reload_action)

        if True in list(map(lambda item: isinstance(item, NoteItem) or isinstance(item, ShapeItem) or isinstance(item, ImageItem), items)):
            duplicate_action = QtGui.QAction("Duplicate", menu)
            duplicate_action.setIcon(QtGui.QIcon(':/icons/new.svg'))
            duplicate_action.triggered.connect(self.duplicateActionSlot)
            menu.addAction(duplicate_action)

        if True in list(map(lambda item: isinstance(item, NoteItem), items)):
            text_edit_action = QtGui.QAction("Text edit", menu)
            text_edit_action.setIcon(QtGui.QIcon(':/icons/show-hostname.svg'))  # TODO: change icon for text edit
            text_edit_action.triggered.connect(self.textEditActionSlot)
            menu.addAction(text_edit_action)

        if True in list(map(lambda item: isinstance(item, ShapeItem), items)):
            style_action = QtGui.QAction("Style", menu)
            style_action.setIcon(QtGui.QIcon(':/icons/drawing.svg'))
            style_action.triggered.connect(self.styleActionSlot)
            menu.addAction(style_action)

        # item must have no parent
        if True in list(map(lambda item: item.parentItem() is None, items)):

            if len(items) > 1:
                horizontal_align_action = QtGui.QAction("Align horizontally", menu)
                horizontal_align_action.setIcon(QtGui.QIcon(':/icons/horizontally.svg'))
                horizontal_align_action.triggered.connect(self.horizontalAlignmentSlot)
                menu.addAction(horizontal_align_action)

                vertical_align_action = QtGui.QAction("Align vertically", menu)
                vertical_align_action.setIcon(QtGui.QIcon(':/icons/vertically.svg'))
                vertical_align_action.triggered.connect(self.verticalAlignmentSlot)
                menu.addAction(vertical_align_action)

            raise_layer_action = QtGui.QAction("Raise one layer", menu)
            raise_layer_action.setIcon(QtGui.QIcon(':/icons/raise_z_value.svg'))
            raise_layer_action.triggered.connect(self.raiseLayerActionSlot)
            menu.addAction(raise_layer_action)

            lower_layer_action = QtGui.QAction("Lower one layer", menu)
            lower_layer_action.setIcon(QtGui.QIcon(':/icons/lower_z_value.svg'))
            lower_layer_action.triggered.connect(self.lowerLayerActionSlot)
            menu.addAction(lower_layer_action)

            delete_action = QtGui.QAction("Delete", menu)
            delete_action.setIcon(QtGui.QIcon(':/icons/delete.svg'))
            delete_action.triggered.connect(self.deleteActionSlot)
            menu.addAction(delete_action)

    def startActionSlot(self):
        """
        Slot to receive events from the start action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and hasattr(item.node(), "start") and item.node().initialized():
                item.node().start()

    def stopActionSlot(self):
        """
        Slot to receive events from the stop action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and hasattr(item.node(), "stop") and item.node().initialized():
                item.node().stop()

    def suspendActionSlot(self):
        """
        Slot to receive events from the suspend action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and hasattr(item.node(), "suspend") and item.node().initialized():
                item.node().suspend()

    def reloadActionSlot(self):
        """
        Slot to receive events from the reload action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and hasattr(item.node(), "reload") and item.node().initialized():
                item.node().reload()

    def configureActionSlot(self):
        """
        Slot to receive events from the configure action in the
        contextual menu.
        """

        items = []
        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and item.node().initialized():
                items.append(item)

        if items:
            self.configureSlot(items)

    def changeHostnameActionSlot(self):
        """
        Slot to receive events from the change hostname action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and item.node().initialized():
                new_hostname, ok = QtGui.QInputDialog.getText(self, "Change hostname", "Hostname:", QtGui.QLineEdit.Normal, item.node().name())
                if ok:
                    if hasattr(item.node(), "validateHostname"):
                        if not item.node().validateHostname(new_hostname):
                            QtGui.QMessageBox.critical(self, "Change hostname", "Invalid name detected for this node: {}".format(new_hostname))
                            continue
                    item.node().update({"name": new_hostname})

    def changeSymbolActionSlot(self):
        """
        Slot to receive events from the change symbol action in the
        contextual menu.
        """

        items = []
        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and item.node().initialized():
                items.append(item)
        if items:
            dialog = SymbolSelectionDialog(self, items)
            dialog.show()
            dialog.exec_()

    def consoleToNode(self, node, aux=False):
        """
        Start a console application to connect to a node.

        :param node: Node instance
        :param aux: auxiliary console mode

        :returns: False if the console application could not be started
        """

        if not hasattr(node, "console") or not node.initialized() or node.status() != Node.started:
            # returns True to ignore this node.
            return True

        if aux and not hasattr(node, "auxConsole"):
            # returns True to ignore this node.
            return True

        if hasattr(node, "serialConsole") and node.serialConsole():
            try:
                from .serial_console import serialConsole
                serialConsole(node.name())
            except (OSError, ValueError) as e:
                QtGui.QMessageBox.critical(self, "Console", "Cannot start serial console application: {}".format(e))
                return False
        else:
            name = node.name()
            if aux:
                console_port = node.auxConsole()
                if console_port is None:
                    QtGui.QMessageBox.critical(self, "Console", "AUX console port not allocated for {}".format(name))
                    return False
            else:
                console_port = node.console()
            console_host = node.server().host
            try:
                from .telnet_console import telnetConsole

                telnet_callback = None

                try:
                    if node.server().isCloud():
                        # override target host with localhost
                        console_host = '127.0.0.1'
                        ep = node.server().tunnel.add_endpoint(console_host, console_port)
                        # override target port with local tunneled port
                        local_addr, _ = ep.get()
                        console_port = local_addr[1]

                        def cb(*args, **kwargs):
                            node.server().tunnel.remove_endpoint(ep)
                            log.debug('Console DONE on port {}'.format(args[2]))
                        telnet_callback = cb

                except AttributeError:
                    pass

                telnetConsole(name, console_host, console_port, telnet_callback)
            except (OSError, ValueError) as e:
                QtGui.QMessageBox.critical(self, "Console", "Cannot start console application: {}".format(e))
                return False
        return True

    def consoleFromItems(self, items):
        """
        Console from scene items.

        :param items: Item instances
        """

        nodes = {}
        for item in items:
            if isinstance(item, NodeItem) and hasattr(item.node(), "console") and item.node().initialized() and item.node().status() == Node.started:
                node = item.node()
                nodes[node.name()] = node

        delay = self._main_window.settings()["delay_console_all"]
        counter = 0
        for name in sorted(nodes.keys()):
            node = nodes[name]
            callback = functools.partial(self.consoleToNode, node)
            self._main_window.run_later(counter, callback)
            counter += delay

    def consoleActionSlot(self):
        """
        Slot to receive events from the console action in the
        contextual menu.
        """

        self.consoleFromItems(self.scene().selectedItems())

    def auxConsoleFromItems(self, items):
        """
        Aux console from scene items.

        :param items: Item instances
        """

        nodes = {}
        for item in items:
            if isinstance(item, NodeItem) and hasattr(item.node(), "auxConsole") and item.node().initialized() and item.node().status() == Node.started:
                node = item.node()
                nodes[node.name()] = node

        delay = self._main_window.settings()["delay_console_all"]
        counter = 0
        for name in sorted(nodes.keys()):
            node = nodes[name]
            callback = functools.partial(self.consoleToNode, node, aux=True)
            self._main_window.run_later(counter, callback)
            counter += delay

    def auxConsoleActionSlot(self):
        """
        Slot to receive events from the auxiliary console action in the
        contextual menu.
        """

        self.auxConsoleFromItems(self.scene().selectedItems())

    def importConfigActionSlot(self):
        """
        Slot to receive events from the import config action in the
        contextual menu.
        """

        items = []
        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and hasattr(item.node(), "importConfig") and item.node().initialized():
                items.append(item)

        if not items:
            return

        if len(items) > 1:
            path = QtGui.QFileDialog.getExistingDirectory(self, "Import directory", ".", QtGui.QFileDialog.ShowDirsOnly)
            if path:
                for item in items:
                    item.node().importConfigFromDirectory(path)
        else:
            item = items[0]
            path, _ = QtGui.QFileDialog.getOpenFileNameAndFilter(self,
                                                                 "Import config",
                                                                 ".",
                                                                 "All files (*.*);;Config files (*.cfg)",
                                                                 "Config files (*.cfg)")
            if path:
                item.node().importConfig(path)
            if hasattr(item.node(), "importPrivateConfig"):
                path, _ = QtGui.QFileDialog.getOpenFileNameAndFilter(self,
                                                                     "Import private-config",
                                                                     ".",
                                                                     "All files (*.*);;Config files (*.cfg)",
                                                                     "Config files (*.cfg)")
                if path:
                    item.node().importPrivateConfig(path)

    def exportConfigActionSlot(self):
        """
        Slot to receive events from the export config action in the
        contextual menu.
        """

        items = []
        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and hasattr(item.node(), "exportConfig") and item.node().initialized():
                items.append(item)

        if not items:
            return

        if len(items) > 1:
            path = QtGui.QFileDialog.getExistingDirectory(self, "Export directory", ".", QtGui.QFileDialog.ShowDirsOnly)
            if path:
                for item in items:
                    item.node().exportConfigToDirectory(path)
        else:
            item = items[0]
            config_path = QtGui.QFileDialog.getSaveFileName(self, "Export config")
            if hasattr(item.node(), "importPrivateConfig"):
                private_config_path = QtGui.QFileDialog.getSaveFileName(self, "Export private-config")
                item.node().exportConfig(config_path, private_config_path)
            else:
                item.node().exportConfig(config_path)

    def saveConfigActionSlot(self):
        """
        Slot to receive events from the save config action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and hasattr(item.node(), "saveConfig") and item.node().initialized():
                item.node().saveConfig()

    def captureActionSlot(self):
        """
        Slot to receive events from the capture action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem) and hasattr(item.node(), "startPacketCapture") and item.node().initialized():
                node = item.node()
                ports = {}
                for port in node.ports():
                    if not port.isFree() and port.packetCaptureSupported() and not port.capturing():
                        for dlt_name, dlt in port.dataLinkTypes().items():
                            key = "Port {} ({} encapsulation: {})".format(port.name(), dlt_name, dlt)
                            ports[key] = [port, dlt]
                if ports:
                    selection, ok = QtGui.QInputDialog.getItem(self, "Capture on {}".format(node.name()), "Please select a port:", list(ports.keys()), 0, False)
                    if ok:
                        if selection in ports:
                            port, dlt = ports[selection]
                            node.startPacketCapture(port, port.captureFileName(node.name()), dlt)
                else:
                    QtGui.QMessageBox.warning(self, "Capture", "No port available for packet capture on {}".format(node.name()))

    def idlepcActionSlot(self):
        """
        Slot to receive events from the idlepc action in the
        contextual menu.
        """

        items = self.scene().selectedItems()
        if len(items) != 1:
            QtGui.QMessageBox.critical(self, "Idle-PC", "Please select only one router")
            return
        item = items[0]
        if isinstance(item, NodeItem) and hasattr(item.node(), "idlepc") and item.node().initialized():
            router = item.node()
            #question = QtGui.QMessageBox.question(self, "Auto Idle-PC", "Would you like to automatically find a suitable Idle-PC value (but not optimal)?", QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)

            #if question == QtGui.QMessageBox.Yes:
            #    router.computeAutoIdlepc(self._autoIdlepcCallback)
            #else:
            router.computeIdlepcs(self._idlepcCallback)

    def _idlepcCallback(self, result, error=False, context={}, **kwargs):
        """
        Slot to allow the user to select an idle-pc value.
        """

        if error:
            QtGui.QMessageBox.critical(self, "Idle-PC", "Error: {}".format(result["message"]))
        else:
            router = context["router"]
            log.info("{} has received Idle-PC proposals".format(router.name()))
            idlepcs = result
            if idlepcs and idlepcs[0] != "0x0":
                dialog = IdlePCDialog(router, idlepcs, parent=self)
                dialog.show()
                dialog.exec_()
            else:
                QtGui.QMessageBox.critical(self, "Idle-PC", "Sorry no Idle-PC values could be computed, please check again with Cisco IOS in a different state")

    def _autoIdlepcCallback(self, result, error=False, context={}, **kwargs):
        """
        Slot to allow the user to select an idlepc value.
        """

        if error:
            QtGui.QMessageBox.critical(self, "Auto Idle-PC", "Error: {}".format(result["message"]))
        else:
            router = context["router"]
            idlepc = result["idlepc"]
            log.info("{} has received the auto idle-pc value: {}".format(router.name(), idlepc))
            router.setIdlepc(idlepc)
            QtGui.QMessageBox.information(self, "Auto Idle-PC", "Idle-PC value {} has been applied on {}".format(idlepc, router.name()))

    def duplicateActionSlot(self):
        """
        Slot to receive events from the duplicate action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if isinstance(item, NoteItem):
                note_item = item.duplicate()
                self.scene().addItem(note_item)
                self._topology.addNote(note_item)
            elif isinstance(item, RectangleItem):
                rectangle_item = item.duplicate()
                self.scene().addItem(rectangle_item)
                self._topology.addRectangle(rectangle_item)
            elif isinstance(item, EllipseItem):
                ellipse_item = item.duplicate()
                self.scene().addItem(ellipse_item)
                self._topology.addEllipse(ellipse_item)
            elif isinstance(item, ImageItem):
                image_item = item.duplicate()
                self.scene().addItem(image_item)
                self._topology.addImage(image_item)

    def styleActionSlot(self):
        """
        Slot to receive events from the style action in the
        contextual menu.
        """

        items = []
        for item in self.scene().selectedItems():
            if isinstance(item, ShapeItem):
                items.append(item)
        if items:
            style_dialog = StyleEditorDialog(self._main_window, items)
            style_dialog.show()
            style_dialog.exec_()

    def textEditActionSlot(self):
        """
        Slot to receive events from the text edit action in the
        contextual menu.
        """

        items = []
        for item in self.scene().selectedItems():
            if isinstance(item, NoteItem):
                items.append(item)
        if items:
            text_edit_dialog = TextEditorDialog(self._main_window, items)
            text_edit_dialog.show()
            text_edit_dialog.exec_()

    def horizontalAlignmentSlot(self):
        """
        Slot to receive events from the horizontal align action in the
        contextual menu.
        """

        horizontal_pos = None
        for item in self.scene().selectedItems():
            if item.parentItem() is None:
                if horizontal_pos is None:
                    horizontal_pos = item.y()
                item.setPos(item.x(), horizontal_pos)

    def verticalAlignmentSlot(self):
        """
        Slot to receive events from the vertical align action in the
        contextual menu.
        """

        vertical_position = None
        for item in self.scene().selectedItems():
            if item.parentItem() is None:
                if vertical_position is None:
                    vertical_position = item.x()
                item.setPos(vertical_position, item.y())

    def raiseLayerActionSlot(self):
        """
        Slot to receive events from the raise one layer action in the
        contextual menu.
        """

        for item in self.scene().selectedItems():
            if item.parentItem() is None:
                current_zvalue = item.zValue()
                item.setZValue(current_zvalue + 1)
                item.update()

    def lowerLayerActionSlot(self):
        """
        Slot to receive events from the lower one layer action in the
        contextual menu.
        """

        show_message = True
        for item in self.scene().selectedItems():
            if item.parentItem() is None:
                current_zvalue = item.zValue()
                item.setZValue(current_zvalue - 1)
                item.update()
                if item.zValue() == -1 and show_message:
                    QtGui.QMessageBox.information(self, "Layer position", "Object moved to a background layer. You will now have to use the right-click action to select this object in the future and raise it to layer 0 to be able to move it")
                    show_message = False

    def deleteActionSlot(self):
        """
        Slot to receive events from the delete action in the
        contextual menu.
        """

        selected_nodes = []
        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem):
                selected_nodes.append(item.node())
        if selected_nodes:
            if len(selected_nodes) > 1:
                question = "Do you want to permanently delete these {} nodes?".format(len(selected_nodes))
            else:
                question = "Do you want to permanently delete {}?".format(selected_nodes[0].name())
            reply = QtGui.QMessageBox.question(self, "Delete", question,
                                               QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)
            if reply == QtGui.QMessageBox.No:
                return
        for item in self.scene().selectedItems():
            if isinstance(item, NodeItem):
                item.node().delete()
                self._topology.removeNode(item.node())
            elif item.parentItem() is None:
                item.delete()

    def createNode(self, node_data, pos):
        """
        Creates a new node on the scene.

        :param node_data: node data to create a new node
        :param pos: position of the drop event

        :returns: NodeItem instance
        """

        try:
            node_module = None
            for module in MODULES:
                instance = module.instance()
                node_class = module.getNodeClass(node_data["class"])
                if node_class in instance.classes():
                    node_module = instance
                    break

            if not node_module:
                raise ModuleError("Could not find any module for {}".format(node_class))

            if "server" not in node_data:
                server = node_module.allocateServer(node_class)
            elif node_data["server"] == "local":
                server = Servers.instance().localServer()
            elif node_data["server"] == "cloud":
                server = Servers.instance().anyCloudServer()
            else:
                try:
                    host, port = node_data["server"].rsplit(":", 1)
                except ValueError:
                    raise ModuleError("Wrong format for server: '{}', please recreate the node in preferences".format(node_data["server"]))
                server = Servers.instance().getRemoteServer(host, port)

            if server is None:
                return

            node = node_module.createNode(node_class, server, self._main_window.project())
            node.error_signal.connect(self._main_window.uiConsoleTextEdit.writeError)
            node.warning_signal.connect(self._main_window.uiConsoleTextEdit.writeWarning)
            node.server_error_signal.connect(self._main_window.uiConsoleTextEdit.writeServerError)
            node_item = NodeItem(node, node_data["default_symbol"], node_data["hover_symbol"])
            node_module.setupNode(node, node_data["name"])
        except ModuleError as e:
            QtGui.QMessageBox.critical(self, "Node creation", "{}".format(e))
            return

        node_item.setPos(self.mapToScene(pos))
        self.scene().addItem(node_item)
        x = node_item.pos().x() - (node_item.boundingRect().width() / 2)
        y = node_item.pos().y() - (node_item.boundingRect().height() / 2)
        node_item.setPos(x, y)
        self._topology.addNode(node)
        self._main_window.uiTopologySummaryTreeWidget.addNode(node)
        return node_item



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/Users/noplay/code/gns3/gns3-gui/gns3/ui/new_project_dialog.ui'
#
# Created: Tue Mar  3 11:41:04 2015
#      by: PyQt4 UI code generator 4.11.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s

try:
    _encoding = QtGui.QApplication.UnicodeUTF8

    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig, _encoding)
except AttributeError:
    def _translate(context, text, disambig):
        return QtGui.QApplication.translate(context, text, disambig)


class Ui_NewProjectDialog(object):

    def setupUi(self, NewProjectDialog):
        NewProjectDialog.setObjectName(_fromUtf8("NewProjectDialog"))
        NewProjectDialog.setWindowModality(QtCore.Qt.ApplicationModal)
        NewProjectDialog.resize(491, 177)
        NewProjectDialog.setModal(True)
        self.gridLayout_2 = QtGui.QGridLayout(NewProjectDialog)
        self.gridLayout_2.setSizeConstraint(QtGui.QLayout.SetFixedSize)
        self.gridLayout_2.setObjectName(_fromUtf8("gridLayout_2"))
        self.uiProjectGroupBox = QtGui.QGroupBox(NewProjectDialog)
        self.uiProjectGroupBox.setObjectName(_fromUtf8("uiProjectGroupBox"))
        self.gridLayout = QtGui.QGridLayout(self.uiProjectGroupBox)
        self.gridLayout.setObjectName(_fromUtf8("gridLayout"))
        self.uiNameLabel = QtGui.QLabel(self.uiProjectGroupBox)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uiNameLabel.sizePolicy().hasHeightForWidth())
        self.uiNameLabel.setSizePolicy(sizePolicy)
        self.uiNameLabel.setObjectName(_fromUtf8("uiNameLabel"))
        self.gridLayout.addWidget(self.uiNameLabel, 0, 0, 1, 1)
        self.uiNameLineEdit = QtGui.QLineEdit(self.uiProjectGroupBox)
        self.uiNameLineEdit.setObjectName(_fromUtf8("uiNameLineEdit"))
        self.gridLayout.addWidget(self.uiNameLineEdit, 0, 1, 1, 4)
        self.uiLocationLabel = QtGui.QLabel(self.uiProjectGroupBox)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uiLocationLabel.sizePolicy().hasHeightForWidth())
        self.uiLocationLabel.setSizePolicy(sizePolicy)
        self.uiLocationLabel.setObjectName(_fromUtf8("uiLocationLabel"))
        self.gridLayout.addWidget(self.uiLocationLabel, 1, 0, 1, 1)
        self.uiLocationLineEdit = QtGui.QLineEdit(self.uiProjectGroupBox)
        self.uiLocationLineEdit.setObjectName(_fromUtf8("uiLocationLineEdit"))
        self.gridLayout.addWidget(self.uiLocationLineEdit, 1, 1, 1, 3)
        self.uiLocationBrowserToolButton = QtGui.QToolButton(self.uiProjectGroupBox)
        self.uiLocationBrowserToolButton.setObjectName(_fromUtf8("uiLocationBrowserToolButton"))
        self.gridLayout.addWidget(self.uiLocationBrowserToolButton, 1, 4, 1, 1)
        self.uiTypeLabel = QtGui.QLabel(self.uiProjectGroupBox)
        self.uiTypeLabel.setEnabled(True)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uiTypeLabel.sizePolicy().hasHeightForWidth())
        self.uiTypeLabel.setSizePolicy(sizePolicy)
        self.uiTypeLabel.setObjectName(_fromUtf8("uiTypeLabel"))
        self.gridLayout.addWidget(self.uiTypeLabel, 2, 0, 1, 1)
        self.uiLocalRadioButton = QtGui.QRadioButton(self.uiProjectGroupBox)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Fixed, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uiLocalRadioButton.sizePolicy().hasHeightForWidth())
        self.uiLocalRadioButton.setSizePolicy(sizePolicy)
        self.uiLocalRadioButton.setChecked(True)
        self.uiLocalRadioButton.setObjectName(_fromUtf8("uiLocalRadioButton"))
        self.gridLayout.addWidget(self.uiLocalRadioButton, 2, 1, 1, 1)
        self.uiCloudRadioButton = QtGui.QRadioButton(self.uiProjectGroupBox)
        self.uiCloudRadioButton.setEnabled(False)
        self.uiCloudRadioButton.setObjectName(_fromUtf8("uiCloudRadioButton"))
        self.gridLayout.addWidget(self.uiCloudRadioButton, 2, 2, 1, 1)
        spacerItem = QtGui.QSpacerItem(201, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 2, 3, 1, 2)
        self.gridLayout_2.addWidget(self.uiProjectGroupBox, 0, 0, 1, 3)
        self.uiOpenProjectPushButton = QtGui.QPushButton(NewProjectDialog)
        self.uiOpenProjectPushButton.setObjectName(_fromUtf8("uiOpenProjectPushButton"))
        self.gridLayout_2.addWidget(self.uiOpenProjectPushButton, 1, 0, 1, 1)
        self.uiRecentProjectsPushButton = QtGui.QPushButton(NewProjectDialog)
        self.uiRecentProjectsPushButton.setObjectName(_fromUtf8("uiRecentProjectsPushButton"))
        self.gridLayout_2.addWidget(self.uiRecentProjectsPushButton, 1, 1, 1, 1)
        self.uiButtonBox = QtGui.QDialogButtonBox(NewProjectDialog)
        self.uiButtonBox.setOrientation(QtCore.Qt.Horizontal)
        self.uiButtonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel | QtGui.QDialogButtonBox.Ok)
        self.uiButtonBox.setObjectName(_fromUtf8("uiButtonBox"))
        self.gridLayout_2.addWidget(self.uiButtonBox, 1, 2, 1, 1)

        self.retranslateUi(NewProjectDialog)
        QtCore.QObject.connect(self.uiButtonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), NewProjectDialog.accept)
        QtCore.QObject.connect(self.uiButtonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), NewProjectDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(NewProjectDialog)

    def retranslateUi(self, NewProjectDialog):
        NewProjectDialog.setWindowTitle(_translate("NewProjectDialog", "New project", None))
        self.uiProjectGroupBox.setTitle(_translate("NewProjectDialog", "Project", None))
        self.uiNameLabel.setText(_translate("NewProjectDialog", "Name:", None))
        self.uiLocationLabel.setText(_translate("NewProjectDialog", "Location:", None))
        self.uiLocationBrowserToolButton.setText(_translate("NewProjectDialog", "Browse...", None))
        self.uiTypeLabel.setText(_translate("NewProjectDialog", "Type:", None))
        self.uiLocalRadioButton.setText(_translate("NewProjectDialog", "Local", None))
        self.uiCloudRadioButton.setText(_translate("NewProjectDialog", "Cloud", None))
        self.uiOpenProjectPushButton.setText(_translate("NewProjectDialog", "&Open a project", None))
        self.uiRecentProjectsPushButton.setText(_translate("NewProjectDialog", "&Recent projects...", None))
